@@ -8,6 +8,7 @@ import inspect
 import os
 import py
 import sys
+import traceback
 
 from requests import Response, exceptions
 from devpi_common.types import cached_property
@@ -23,6 +24,7 @@ from . import __version__ as server_version
 class Fatal(Exception):
     pass
 
+
 def fatal(msg):
     raise Fatal(msg)
 
@@ -39,7 +41,7 @@ def check_compatible_version(config):
         if state_ver[0] != DATABASE_VERSION:
             fatal("Incompatible state: server %s cannot run serverdir "
                   "%s created at database version %s.\n"
-                  "Use devpi-export (or --export) from older version, then "
+                  "Use devpi-export from older version, then "
                   "devpi-import with newer version."
                   % (server_version, config.serverdir, state_ver[0]))
 
@@ -60,22 +62,8 @@ def set_state_version(config, version=DATABASE_VERSION):
     versionfile.write(version)
 
 
-def check_python_version():
-    if sys.version_info.major < 3:
-        import warnings
-        warnings.warn(
-            "DEPRECATION: Python 2.7 will reach the end of its life on "
-            "January 1st, 2020. Please upgrade your Python as Python 2.7 "
-            "won't be maintained after that date.\n"
-            "The recommended way for upgrading devpi-server to Python 3.x "
-            "is to first use the --export option with your existing "
-            "Python 2.7 setup. Then use --import in a new Python 3.x "
-            "environment.")
-
-
 def main(argv=None):
     """ devpi-server command line entry point. """
-    check_python_version()
     pluginmanager = get_pluginmanager()
     try:
         return _main(pluginmanager, argv=argv)
@@ -85,18 +73,18 @@ def main(argv=None):
         return 1
 
 
-def xom_from_config(config):
+def xom_from_config(config, init=False):
     check_compatible_version(config)
 
     # read/create node UUID and role of this server
     config.init_nodeinfo()
 
-    if config.sqlite_file_needed_but_missing():
+    if not init and config.sqlite_file_needed_but_missing():
         fatal(
-            "No sqlite storage found in serverdir."
+            "No sqlite storage found in %s."
             " Or you need to run with --storage to specify the storage type,"
             " or you first need to run devpi-init or devpi-import"
-            " in order to create the sqlite database."
+            " in order to create the sqlite database." % config.serverdir
         )
 
     return XOM(config)
@@ -125,69 +113,14 @@ def _main(pluginmanager, argv=None):
         print(server_version)
         return 0
 
-    if args.genconfig:
-        from devpi_server.genconfig import genconfig
-        import warnings
-        warnings.warn(
-            "DEPRECATION: the --gen-config option is deprecated, use the "
-            "devpi-gen-config command instead")
-        return genconfig(config, argv)
-
     # now we can configure logging
     configure_logging(config.args)
 
-    if args.init:
-        import warnings
-        warnings.warn(
-            "DEPRECATION: the --init option is deprecated, use the "
-            "devpi-init command instead")
-        if config.path_nodeinfo.exists():
-            fatal("The path '%s' already contains devpi-server data." % config.serverdir)
-    elif not args.import_:
-        if not config.path_nodeinfo.exists():
-            fatal("The path '%s' contains no devpi-server data, use devpi-init to initialize." % config.serverdir)
-
-    if args.init or args.import_:
-        sdir = config.serverdir
-        if not (sdir.exists() and len(sdir.listdir()) >= 2):
-            set_state_version(config, DATABASE_VERSION)
+    if not config.path_nodeinfo.exists():
+        fatal("The path '%s' contains no devpi-server data, use devpi-init to initialize." % config.serverdir)
 
     xom = xom_from_config(config)
 
-    if args.init:
-        init_default_indexes(xom)
-
-    if args.start or args.stop or args.log or args.status:
-        xprocdir = config.serverdir.join(".xproc")
-        from devpi_server.bgserver import BackgroundServer
-        tw = py.io.TerminalWriter()
-        bgserver = BackgroundServer(tw, xprocdir)
-        if args.start:
-            return bgserver.start(args, argv[1:])
-        elif args.stop:
-            return bgserver.stop()
-        elif args.log:
-            return bgserver.log()
-        elif args.status:
-            if bgserver.info.isrunning():
-                bgserver.line("server is running with pid %s" %
-                              bgserver.info.pid)
-            else:
-                bgserver.line("no server is running")
-            return
-
-    if args.passwd:
-        from devpi_server.model import run_passwd
-        import warnings
-        warnings.warn(
-            "DEPRECATION: the --passwd option is deprecated, use the "
-            "devpi-passwd command instead")
-        with xom.keyfs.transaction(write=True):
-            return run_passwd(xom.model, config.args.passwd)
-
-    if args.init:
-        return 0
-
     return xom.main()
 
 
@@ -196,32 +129,19 @@ def make_application():
     config = parseoptions(get_pluginmanager(), [])
     return XOM(config).create_app()
 
+
 def wsgi_run(xom, app):
     from waitress import serve
-    host = xom.config.args.host
-    port = xom.config.args.port
-    unix_socket = xom.config.args.unix_socket
-    kwargs = dict(
-        threads=xom.config.args.threads,
-        max_request_body_size=xom.config.args.max_request_body_size)
-    if unix_socket is not None:
-        kwargs['unix_socket'] = unix_socket
-        if xom.config.args.unix_socket_perms is not None:
-            kwargs['unix_socket_perms'] = xom.config.args.unix_socket_perms
-        if host == 'localhost':
-            host = None
-        if port == 3141:
-            port = None
-    if host or port:
-        kwargs['host'] = host
-        kwargs['port'] = port
     log = xom.log
+    kwargs = xom.config.waitress_info["kwargs"]
+    addresses = xom.config.waitress_info["addresses"]
     log.info("devpi-server version: %s", server_version)
     log.info("serverdir: %s" % xom.config.serverdir)
     log.info("uuid: %s" % xom.config.nodeinfo["uuid"])
-    hostaddr = "http://%s:%s" % (host, port)
-    hostaddr6 = "http://[%s]:%s" % (host, port)
-    log.info("serving at url: %s (might be %s for IPv6)", hostaddr, hostaddr6)
+    if len(addresses) == 1:
+        log.info("serving at url: %s", addresses[0])
+    else:
+        log.info("serving at urls: %s", ", ".join(addresses))
     log.info("using %s threads", kwargs['threads'])
     log.info("bug tracker: https://github.com/devpi/devpi/issues")
     log.info("IRC: #devpi on irc.freenode.net")
@@ -259,6 +179,21 @@ class XOM:
         self.log = threadlog
         self.polling_replicas = {}
         self._stagecache = {}
+        if self.is_replica():
+            from devpi_server.replica import ReplicaThread
+            from devpi_server.replica import register_key_subscribers
+            search_path = self.config.replica_file_search_path
+            if search_path and not os.path.exists(search_path):
+                fatal(
+                    "search path for existing replica files doesn't "
+                    "exist: %s" % search_path)
+            register_key_subscribers(self)
+            # the replica thread replays keyfs changes
+            # and project-specific changes are discovered
+            # and replayed through the PypiProjectChange event
+            if not self.config.requests_only:
+                self.replica_thread = ReplicaThread(self)
+                self.thread_pool.register(self.replica_thread)
 
     def get_singleton(self, indexpath, key):
         """ return a per-xom singleton for the given indexpath and key
@@ -292,27 +227,6 @@ class XOM:
 
     def main(self):
         xom = self
-        args = xom.config.args
-        if args.export:
-            from devpi_server.importexport import do_export
-            import warnings
-            warnings.warn(
-                "DEPRECATION: the --export option is deprecated, use the "
-                "devpi-export command instead")
-            #xom.thread_pool.start_one(xom.keyfs.notifier)
-            return do_export(args.export, xom)
-
-        if args.import_:
-            from devpi_server.importexport import do_import
-            import warnings
-            warnings.warn(
-                "DEPRECATION: the --import option is deprecated, use the "
-                "devpi-import command instead")
-            # we need to start the keyfs notifier so that import
-            # can wait on completion of events
-            if args.wait_for_events:
-                xom.thread_pool.start_one(xom.keyfs.notifier)
-            return do_import(args.import_, xom)
 
         # creation of app will register handlers of key change events
         # which cannot happen anymore after the tx notifier has started
@@ -377,22 +291,28 @@ class XOM:
             headers.update(extra_headers)
         try:
             resp = self._httpsession.get(
-                        url, stream=True,
-                        allow_redirects=allow_redirects,
-                        headers=headers,
-                        timeout=timeout or self.config.args.request_timeout)
+                url, stream=True,
+                allow_redirects=allow_redirects,
+                headers=headers,
+                timeout=timeout or self.config.args.request_timeout)
             return resp
-        except OSError:
+        except OSError as e:
             location = get_caller_location()
-            threadlog.warn("OS error during httpget of %s at %s", url, location)
+            threadlog.warn(
+                "OS error during httpget of %s at %s: %s",
+                url, location, traceback.format_exception_only(e.__class__, e))
             return FatalResponse(url, repr(sys.exc_info()[1]))
-        except exceptions.ConnectionError:
+        except exceptions.ConnectionError as e:
             location = get_caller_location()
-            threadlog.warn("Connection error during httpget of %s at %s", url, location)
+            threadlog.warn(
+                "Connection error during httpget of %s at %s: %s",
+                url, location, traceback.format_exception_only(e.__class__, e))
             return FatalResponse(url, repr(sys.exc_info()[1]))
-        except self._httpsession.Errors:
+        except self._httpsession.Errors as e:
             location = get_caller_location()
-            threadlog.warn("HTTPError during httpget of %s at %s", url, location)
+            threadlog.warn(
+                "HTTPError during httpget of %s at %s: %s",
+                url, location, traceback.format_exception_only(e.__class__, e))
             return FatalResponse(url, repr(sys.exc_info()[1]))
 
     def view_deriver(self, view, info):
@@ -416,27 +336,24 @@ class XOM:
     view_deriver.options = ('is_mutating',)
 
     def create_app(self):
-        from devpi_server.view_auth import DevpiAuthenticationPolicy
+        from devpi_server.view_auth import DevpiSecurityPolicy
         from devpi_server.views import ContentTypePredicate
         from devpi_server.views import OutsideURLMiddleware
         from devpi_server.views import route_url, INSTALLER_USER_AGENT
         from pkg_resources import get_distribution
-        from pyramid.authorization import ACLAuthorizationPolicy
         from pyramid.config import Configurator
         from pyramid.viewderivers import INGRESS
         log = self.log
         log.debug("creating application in process %s", os.getpid())
         pyramid_config = Configurator(root_factory='devpi_server.view_auth.RootFactory')
-        pyramid_config.set_authentication_policy(DevpiAuthenticationPolicy(self))
-        pyramid_config.set_authorization_policy(ACLAuthorizationPolicy())
+        pyramid_config.set_security_policy(DevpiSecurityPolicy(self))
 
         version_info = [
             ("devpi-server", get_distribution("devpi_server").version)]
         for plug, distinfo in self.config.pluginmanager.list_plugin_distinfo():
-            threadlog.info("Found plugin %s-%s (%s)." % (
-                distinfo.project_name, distinfo.version, distinfo.location))
             key = (distinfo.project_name, distinfo.version)
             if key not in version_info:
+                threadlog.info("Found plugin %s-%s." % key)
                 version_info.append(key)
         version_info.sort()
         pyramid_config.registry['devpi_version_info'] = version_info
@@ -457,19 +374,17 @@ class XOM:
                             "%s.%s" % (x.__module__, x.__name__)
                             for x in ixclasses)))
         self.config.hook.devpiserver_pyramid_configure(
-                config=self.config,
-                pyramid_config=pyramid_config)
+            config=self.config,
+            pyramid_config=pyramid_config)
 
         pyramid_config.add_view_deriver(self.view_deriver, under=INGRESS)
         pyramid_config.add_view_predicate('content_type', ContentTypePredicate)
 
-        pyramid_config.add_route("/+changelog/{serial}",
-                                 r"/+changelog/{serial:\d+}")
-        pyramid_config.add_route("/+changelog/{serial}-",
-                                 r"/+changelog/{serial:\d+}-")
         pyramid_config.add_route("/+status", "/+status")
         pyramid_config.add_route("/+api", "/+api", accept="application/json")
-        pyramid_config.add_route("{path:.*}/+api", "{path:.*}/+api", accept="application/json")
+        pyramid_config.add_route("/{user}/+api", "/{user}/+api", accept="application/json")
+        pyramid_config.add_route("/{user}/{index}/+api", "/{user}/{index}/+api", accept="application/json")
+        pyramid_config.add_route("/+authcheck", "/+authcheck")
         pyramid_config.add_route("/+login", "/+login", accept="application/json")
         pyramid_config.add_route("/{user}/{index}/+e/{relpath:.*}", "/{user}/{index}/+e/{relpath:.*}")
         pyramid_config.add_route("/{user}/{index}/+f/{relpath:.*}", "/{user}/{index}/+f/{relpath:.*}")
@@ -484,10 +399,13 @@ class XOM:
         pyramid_config.add_route("/{user}/{index}/{project}/{version}",
                                  "/{user}/{index}/{project}/{version:[^/]+/?}")
         pyramid_config.add_route(
-            "simple_redirect", "/{user}/{index}/{project:[^/]+/?}",
+            "installer_simple", "/{user}/{index:[^/]+/?}",
             header="User-Agent:" + INSTALLER_USER_AGENT,
-            accept="text/html",
-        )
+            accept="text/html", request_method="GET")
+        pyramid_config.add_route(
+            "installer_simple_project", "/{user}/{index}/{project:[^/]+/?}",
+            header="User-Agent:" + INSTALLER_USER_AGENT,
+            accept="text/html", request_method="GET")
         pyramid_config.add_route("/{user}/{index}/{project}",
                                  "/{user}/{index}/{project:[^/]+/?}")
         pyramid_config.add_route("/{user}/{index}/", "/{user}/{index}/")
@@ -506,26 +424,13 @@ class XOM:
         pyramid_config.add_request_method(stage_url)
         pyramid_config.add_request_method(simpleindex_url)
         pyramid_config.add_request_method(apifatal)
+        pyramid_config.add_request_method(apireturn_for_request, "apireturn")
 
         # overwrite route_url method with our own
         pyramid_config.add_request_method(route_url)
         # XXX end hack
-        pyramid_config.scan()
+        pyramid_config.scan("devpi_server.views")
         app = pyramid_config.make_wsgi_app()
-        if self.is_replica():
-            from devpi_server.replica import ReplicaThread, register_key_subscribers
-            search_path = self.config.args.replica_file_search_path
-            if search_path and not os.path.exists(search_path):
-                fatal(
-                    "search path for existing replica files doesn't "
-                    "exist: %s" % search_path)
-            register_key_subscribers(self)
-            self.replica_thread = ReplicaThread(self)
-            # the replica thread replays keyfs changes
-            # and project-specific changes are discovered
-            # and replayed through the PypiProjectChange event
-            if not self.config.requests_only:
-                self.thread_pool.register(self.replica_thread)
         return OutsideURLMiddleware(app, self)
 
     def is_master(self):
@@ -580,17 +485,21 @@ def apifatal(request, *args, **kwargs):
     apireturn(*args, **kwargs)
 
 
+def apireturn_for_request(request, *args, **kwargs):
+    apireturn(*args, **kwargs)
+
+
 def set_default_indexes(model):
     root_user = model.get_user("root")
     if not root_user:
         root_user = model.create_user(
             "root",
-            model.xom.config.args.root_passwd,
-            pwhash=model.xom.config.args.root_passwd_hash)
+            model.xom.config.root_passwd,
+            pwhash=model.xom.config.root_passwd_hash)
         threadlog.info("created root user")
     userconfig = root_user.key.get(readonly=False)
     indexes = userconfig["indexes"]
-    if "pypi" not in indexes and not model.xom.config.args.no_root_pypi:
+    if "pypi" not in indexes and not model.xom.config.no_root_pypi:
         indexes["pypi"] = _pypi_ixconfig_default.copy()
         root_user.key.set(userconfig)
         threadlog.info("created root/pypi index")
