@@ -15,6 +15,10 @@ try:
     from itertools import zip_longest
 except ImportError:
     from itertools import izip_longest as zip_longest
+try:
+    from pyramid.authorization import Allow, Authenticated, Everyone
+except ImportError:
+    from pyramid.security import Allow, Authenticated, Everyone
 from time import gmtime, strftime
 from .auth import hash_password, verify_and_update_password_hash
 from .config import hookimpl
@@ -23,12 +27,16 @@ from .log import threadlog, thread_current_log
 from .readonly import get_mutable_deepcopy
 
 
-def join_requires(links, requires_python):
-    # build list of (key, href, require_python) tuples
+notset = object()
+
+
+def join_links_data(links, requires_python, yanked):
+    # build list of (key, href, require_python, yanked) tuples
     result = []
-    for link, require_python in zip_longest(links, requires_python, fillvalue=None):
+    links = zip_longest(links, requires_python, yanked, fillvalue=None)
+    for link, require_python, yanked in links:
         key, href = link
-        result.append((key, href, require_python))
+        result.append((key, href, require_python, yanked))
     return result
 
 
@@ -113,8 +121,24 @@ class RootModel:
         self.xom = xom
         self.keyfs = xom.keyfs
 
-    def create_user(self, username, password, email=None, pwhash=None):
-        return User.create(self, username, password, email, pwhash=pwhash)
+    def create_user(self, username, password, **kwargs):
+        userlist = self.keyfs.USERLIST.get(readonly=False)
+        if username in userlist:
+            raise InvalidUser("username '%s' already exists" % username)
+        if not is_valid_name(username):
+            raise InvalidUser(
+                "username '%s' contains characters that aren't allowed. "
+                "Any ascii symbol besides -.@_ is blocked." % username)
+        user = User(self, username)
+        kwargs.update(created=strftime("%Y-%m-%dT%H:%M:%SZ", gmtime()))
+        user._modify(password=password, **kwargs)
+        userlist.add(username)
+        self.keyfs.USERLIST.set(userlist)
+        if "email" in kwargs:
+            threadlog.info("created user %r with email %r" % (username, kwargs["email"]))
+        else:
+            threadlog.info("created user %r" % username)
+        return user
 
     def get_user(self, name):
         user = User(self, name)
@@ -166,6 +190,11 @@ def ensure_list(data):
     return list(filter(None, (x.strip() for x in data.split(","))))
 
 
+class ACLList(list):
+    # marker class for devpiserver_indexconfig_defaults
+    pass
+
+
 def ensure_acl_list(data):
     data = ensure_list(data)
     for index, name in enumerate(data):
@@ -215,7 +244,31 @@ def is_valid_name(name):
     return not name_char_blocklist_regexp.search(name)
 
 
+class InvalidUserconfig(Exception):
+    def __init__(self, messages):
+        if isinstance(messages, py.builtin._basestring):
+            messages = [messages]
+        self.messages = messages
+        Exception.__init__(self, messages)
+
+
 class User:
+    # ignored_keys are skipped on create and modify
+    ignored_keys = frozenset(('indexes', 'username'))
+    # info keys are updated on create and modify and input is ignored
+    info_keys = frozenset(('created', 'modified'))
+    hidden_keys = frozenset((
+        "password", "pwhash", "pwsalt"))
+    public_keys = frozenset((
+        "custom_data", "description", "email", "title"))
+    # allowed_keys can be modified
+    allowed_keys = hidden_keys.union(public_keys)
+    known_keys = allowed_keys.union(ignored_keys, info_keys)
+    # modification update keys control which changed keys will trigger
+    # an update of the modified key
+    modification_update_keys = known_keys - info_keys.union(ignored_keys)
+    # visible_keys are returned via json
+    visible_keys = ignored_keys.union(info_keys, public_keys)
 
     def __init__(self, parent, name):
         self.__parent__ = parent
@@ -227,52 +280,60 @@ class User:
     def key(self):
         return self.keyfs.USER(user=self.name)
 
-    @classmethod
-    def create(cls, model, username, password, email, pwhash=None):
-        userlist = model.keyfs.USERLIST.get(readonly=False)
-        if username in userlist:
-            raise InvalidUser("username '%s' already exists" % username)
-        if not is_valid_name(username):
-            raise InvalidUser(
-                "username '%s' contains characters that aren't allowed. "
-                "Any ascii symbol besides -.@_ is blocked." % username)
-        user = cls(model, username)
-        with user.key.update() as userconfig:
-            if password is not None or pwhash:
-                user._setpassword(userconfig, password, pwhash=pwhash)
-            if email:
-                userconfig["email"] = email
-            userconfig.setdefault("indexes", {})
-        userlist.add(username)
-        model.keyfs.USERLIST.set(userlist)
-        threadlog.info("created user %r with email %r" %(username, email))
-        return user
+    def get_cleaned_config(self, **kwargs):
+        result = {}
+        for key in self.allowed_keys:
+            if key not in kwargs:
+                continue
+            result[key] = kwargs[key]
+        return result
+
+    def validate_config(self, **kwargs):
+        unknown_keys = set(kwargs) - self.known_keys
+        if unknown_keys:
+            raise InvalidUserconfig(
+                "Unknown keys in user config: %s" % ", ".join(unknown_keys))
 
     def _set(self, newuserconfig):
         with self.key.update() as userconfig:
             userconfig.update(newuserconfig)
             threadlog.info("internal: set user information %r", self.name)
 
-    def modify(self, password=None, **kwargs):
+    def _modify(self, password=None, pwhash=None, **kwargs):
+        self.validate_config(**kwargs)
+        modified = {}
         with self.key.update() as userconfig:
-            modified = []
-            if password is not None:
-                self._setpassword(userconfig, password)
-                modified.append("password=*******")
+            if password is not None or pwhash:
+                self._setpassword(userconfig, password, pwhash=pwhash)
+                modified["password"] = "*******"
                 kwargs['pwsalt'] = None
             for key, value in kwargs.items():
                 key = ensure_unicode(key)
-                if key == 'username':
-                    continue
                 if value:
-                    userconfig[key] = value
+                    if userconfig.get(key, notset) != value:
+                        userconfig[key] = value
+                        if key in self.hidden_keys:
+                            value = "*******"
+                        modified[key] = value
                 elif key in userconfig:
                     del userconfig[key]
-                if key in ('pwsalt', 'pwhash') and value:
-                    value = "*******"
-                modified.append("%s=%s" % (key, value))
-            threadlog.info("modified user %r: %s", self.name,
-                           ", ".join(modified))
+                    modified[key] = None
+            userconfig.setdefault("indexes", {})
+            if self.modification_update_keys.intersection(modified):
+                if "created" not in userconfig:
+                    # old data will be set to epoch
+                    modified["created"] = userconfig["created"] = "1970-01-01T00:00:00Z"
+                if "created" not in kwargs:
+                    # only set modified if not created at the same time
+                    modified_ts = strftime("%Y-%m-%dT%H:%M:%SZ", gmtime())
+                    modified["modified"] = userconfig["modified"] = modified_ts
+        return ["%s=%s" % (k, v) for k, v in sorted(modified.items())]
+
+    def modify(self, **kwargs):
+        kwargs = self.get_cleaned_config(**kwargs)
+        modified = self._modify(**kwargs)
+        threadlog.info("modified user %r: %s", self.name,
+                       ", ".join(modified))
 
     def _setpassword(self, userconfig, password, pwhash=None):
         if pwhash:
@@ -299,7 +360,7 @@ class User:
         pwhash = userconfig["pwhash"]
         valid, newhash = verify_and_update_password_hash(password, pwhash, salt)
         if valid:
-            if newhash:
+            if newhash and self.keyfs.tx.write:
                 self.modify(pwsalt=None, pwhash=newhash)
             return True
         return False
@@ -309,8 +370,9 @@ class User:
         if not d:
             return d
         if not credentials:
-            d.pop("pwsalt", None)
-            d.pop("pwhash", None)
+            for key in list(d):
+                if key not in self.visible_keys:
+                    del d[key]
         d["username"] = self.name
         return d
 
@@ -324,6 +386,12 @@ class User:
         stage = self._getstage(index, type, {"type": type})
         if isinstance(stage.customizer, UnknownCustomizer):
             raise InvalidIndexconfig("unknown index type %r" % type)
+        # apply default values for new indexes
+        for key, value in stage.get_default_config_items():
+            kwargs.setdefault(key, value)
+        # apply default values from customizer class for new indexes
+        for key, value in stage.customizer.get_default_config_items():
+            kwargs.setdefault(key, value)
         stage._modify(**kwargs)
         threadlog.info("created index %s: %s", stage.name, stage.ixconfig)
         return stage
@@ -347,6 +415,12 @@ class User:
             return None
         return self._getstage(indexname, ixconfig["type"], ixconfig)
 
+    def getstages(self):
+        stages = []
+        for index in self.get()["indexes"]:
+            stages.append(self.getstage(index))
+        return stages
+
 
 class InvalidIndexconfig(Exception):
     def __init__(self, messages):
@@ -357,7 +431,6 @@ class InvalidIndexconfig(Exception):
 
 
 def get_principals(value):
-    from pyramid.security import Authenticated, Everyone
     principals = set(value)
     if ':AUTHENTICATED:' in principals:
         principals.remove(':AUTHENTICATED:')
@@ -373,12 +446,30 @@ class BaseStageCustomizer(object):
 
     def __init__(self, stage):
         self.stage = stage
+        self.hooks = self.stage.xom.config.hook
 
     # get_principals_for_* methods for each of the following permissions:
     # pypi_submit, toxresult_upload, index_delete, index_modify,
     # del_entry, del_project, del_verdata
     # also see __acl__ method of BaseStage
 
+    def get_principals_for_pkg_read(self, restrict_modify=None):
+        principals = self.hooks.devpiserver_stage_get_principals_for_pkg_read(
+            ixconfig=self.stage.ixconfig)
+        if principals is None:
+            principals = {':ANONYMOUS:'}
+        else:
+            principals = set(principals)
+        if self.stage.xom.is_master():
+            # replicas always need to be able to download packages
+            principals.add("+replica")
+        # admins should always be able to read the packages
+        if restrict_modify is None:
+            principals.add("root")
+        else:
+            principals.update(restrict_modify)
+        return principals
+
     def get_principals_for_pypi_submit(self, restrict_modify=None):
         return self.stage.ixconfig.get("acl_upload", [])
 
@@ -427,7 +518,7 @@ class BaseStageCustomizer(object):
             Will only be called for custom options, not for existing options
             of a private index.
             """
-        return ()
+        pass
 
     def validate_config(self, oldconfig, newconfig):
         """ Validates the index config.
@@ -464,7 +555,8 @@ class BaseStageCustomizer(object):
         """ Called when a list of simple links is returned.
 
             Returns None for no filtering, or an iterator returning
-            True for items to keep and False for items to remove."""
+            True for items to keep and False for items to remove.
+            The size of the tuples in links might grow, develop defensively."""
         return None
 
 
@@ -521,13 +613,10 @@ class BaseStage(object):
                 "The stage customizer for '%s' defines keys which conflict "
                 "with existing index configuration keys: %s"
                 % (index_type, ", ".join(sorted(conflicting))))
-        # get default values from the stage class
+        # prevent default values from being removed
         for key, value in self.get_default_config_items():
             if kwargs.get(key) is RemoveValue:
                 raise InvalidIndexconfig("Default values can't be removed.")
-            # if current ixconfig already has the value, use that
-            value = self.ixconfig.get(key, value)
-            kwargs.setdefault(key, value)
         # now process any key known by the stage class
         for key in stage_keys:
             if key not in kwargs:
@@ -540,13 +629,10 @@ class BaseStage(object):
                         "The key '%s' wasn't processed."
                         % (key))
             ixconfig[key] = value
-        # next get defaults from the customizer class
+        # prevent removal of defaults from the customizer class
         for key, value in self.customizer.get_default_config_items():
             if kwargs.get(key) is RemoveValue:
                 raise InvalidIndexconfig("Default values can't be removed.")
-            # if current ixconfig already has the value, use that
-            value = self.ixconfig.get(key, value)
-            kwargs.setdefault(key, value)
         # and process any key known by the customizer class
         for key in customizer_keys:
             if key not in kwargs:
@@ -569,7 +655,14 @@ class BaseStage(object):
                     "existing index configuration keys for '%s': %s"
                     % (index_type, ", ".join(sorted(conflicting))))
             for key, value in defaults.items():
-                ixconfig.setdefault(key, kwargs.pop(key, value))
+                new_value = kwargs.pop(key, value)
+                if isinstance(value, bool):
+                    new_value = ensure_boolean(new_value)
+                elif isinstance(value, ACLList):
+                    new_value = ensure_acl_list(new_value)
+                elif isinstance(value, (list, tuple, set)):
+                    new_value = ensure_list(new_value)
+                ixconfig.setdefault(key, new_value)
         # XXX backward compatibility for old exports where these could appear
         # on mirror indexes
         # and pypi_whitelist also still needs to be here for existing dbs
@@ -614,21 +707,21 @@ class BaseStage(object):
         # compatibility access method used by devpi-web and tests
         project = normalize_name(project)
         try:
-            return [self._make_elink(project, key, href, require_python)
-                    for key, href, require_python in self.get_simplelinks(project)]
+            return [self._make_elink(project, link_info)
+                    for link_info in self.get_simplelinks(project)]
         except self.UpstreamNotFoundError:
             return []
 
     def get_releaselinks_perstage(self, project):
         # compatibility access method for devpi-findlinks and possibly other plugins
         project = normalize_name(project)
-        return [self._make_elink(project, key, href, require_python)
-                for key, href, require_python in self.get_simplelinks_perstage(project)]
+        return [self._make_elink(project, link_info)
+                for link_info in self.get_simplelinks_perstage(project)]
 
-    def _make_elink(self, project, key, href, require_python):
-        rp = SimplelinkMeta((key, href, require_python))
+    def _make_elink(self, project, link_info):
+        rp = SimplelinkMeta(link_info)
         linkdict = {"entrypath": rp._url.path, "hash_spec": rp._url.hash_spec,
-                    "require_python": require_python}
+                    "require_python": rp.require_python, "yanked": rp.yanked}
         return ELink(self.filestore, linkdict, project, rp.version)
 
     def get_linkstore_perstage(self, name, version, readonly=True):
@@ -758,32 +851,58 @@ class BaseStage(object):
                 iterator = self.customizer.get_simple_links_filter_iter(project, res)
                 if iterator is not None:
                     res = apply_filter_iter(res, iterator)
-                for key, href, require_python in res:
+                for link_info in res:
+                    key = link_info[0]
                     if key not in seen:
                         seen.add(key)
-                        all_links.append((key, href, require_python))
+                        all_links.append(link_info)
         except self.UpstreamNotFoundError:
             return []
 
         if sorted_links:
-            all_links = [(v.key, v.href, v.require_python)
+            all_links = [(v.key, v.href, v.require_python, v.yanked)
                         for v in sorted(map(SimplelinkMeta, all_links), reverse=True)]
         return all_links
 
+    def get_whitelist_inheritance(self):
+        return self.ixconfig.get("mirror_whitelist_inheritance", "union")
+
     def get_mirror_whitelist_info(self, project):
         project = ensure_unicode(project)
         private_hit = whitelisted = False
+        whitelist_inheritance = self.get_whitelist_inheritance()
+        whitelist = None
         for stage in self.sro():
-            in_index = stage.has_project_perstage(project)
             if stage.ixconfig["type"] == "mirror":
+                if private_hit and not whitelisted:
+                    # don't check the mirror for private packages
+                    return dict(
+                        has_mirror_base=False,
+                        blocked_by_mirror_whitelist=stage.name)
+                in_index = stage.has_project_perstage(project)
                 has_mirror_base = in_index and (not private_hit or whitelisted)
                 blocked_by_mirror_whitelist = in_index and private_hit and not whitelisted
                 return dict(
                     has_mirror_base=has_mirror_base,
                     blocked_by_mirror_whitelist=stage.name if blocked_by_mirror_whitelist else None)
+            else:
+                in_index = stage.has_project_perstage(project)
             private_hit = private_hit or in_index
-            whitelist = set(stage.ixconfig.get("mirror_whitelist", set()))
-            whitelisted = whitelisted or '*' in whitelist or project in whitelist
+            stage_whitelist = set(
+                stage.ixconfig.get("mirror_whitelist", set()))
+            if whitelist is None:
+                whitelist = stage_whitelist
+            else:
+                if whitelist_inheritance == "union":
+                    whitelist = whitelist.union(stage_whitelist)
+                elif whitelist_inheritance == "intersection":
+                    whitelist = whitelist.intersection(stage_whitelist)
+                else:
+                    raise RuntimeError(
+                        "Unknown whitelist_inheritance setting '%s'"
+                        % whitelist_inheritance)
+            if whitelisted or whitelist.intersection(('*', project)):
+                whitelisted = True
         return dict(
             has_mirror_base=False,
             blocked_by_mirror_whitelist=None)
@@ -834,8 +953,6 @@ class BaseStage(object):
             return newconfig
 
     def modify(self, index=None, **kw):
-        if self.customizer.readonly:
-            raise ReadonlyIndex("index is marked read only")
         newconfig = self._modify(**kw)
         threadlog.info("modified index %s: %s", self.name, newconfig)
         return newconfig
@@ -853,6 +970,10 @@ class BaseStage(object):
         if not self.filter_projects([project]):
             return
         whitelisted = private_hit = False
+        # the default value if the setting is missing is the old behaviour,
+        # so existing indexes work as before
+        whitelist_inheritance = self.get_whitelist_inheritance()
+        whitelist = None
         for stage in self.sro():
             if stage.ixconfig["type"] == "mirror":
                 if private_hit:
@@ -863,8 +984,20 @@ class BaseStage(object):
                     threadlog.debug("private package %r whitelisted at stage %s",
                                     project, whitelisted.name)
             else:
-                whitelist = set(stage.ixconfig.get("mirror_whitelist", set()))
-                if '*' in whitelist or project in whitelist:
+                stage_whitelist = set(
+                    stage.ixconfig.get("mirror_whitelist", set()))
+                if whitelist is None:
+                    whitelist = stage_whitelist
+                else:
+                    if whitelist_inheritance == "union":
+                        whitelist = whitelist.union(stage_whitelist)
+                    elif whitelist_inheritance == "intersection":
+                        whitelist = whitelist.intersection(stage_whitelist)
+                    else:
+                        raise RuntimeError(
+                            "Unknown whitelist_inheritance setting '%s'"
+                            % whitelist_inheritance)
+                if whitelist.intersection(('*', project)):
                     whitelisted = stage
                 elif stage.has_project_perstage(project):
                     private_hit = True
@@ -873,7 +1006,6 @@ class BaseStage(object):
                 if not stage.has_project_perstage(project):
                     continue
                 res = getattr(stage, opname)(**kw)
-                private_hit = private_hit or res
                 yield stage, res
             except self.UpstreamError as exc:
                 # If we are currently checking ourself raise the error, it is fatal
@@ -886,6 +1018,7 @@ class BaseStage(object):
         todo = [self]
         todo_mirrors = []
         seen = set()
+        devpiserver_sro_skip = self.xom.config.hook.devpiserver_sro_skip
         while todo:
             stage = todo.pop(0)
             yield stage
@@ -897,6 +1030,11 @@ class BaseStage(object):
                         "Index %s refers to non-existing base %s.",
                         self.name, base)
                     continue
+                if devpiserver_sro_skip(stage=self, base_stage=current_stage):
+                    threadlog.warn(
+                        "Index %s base %s excluded via devpiserver_sro_skip.",
+                        self.name, base)
+                    continue
                 if base not in seen:
                     if current_stage.ixconfig['type'] == 'mirror':
                         todo_mirrors.append(current_stage)
@@ -906,9 +1044,8 @@ class BaseStage(object):
             yield stage
 
     def __acl__(self):
-        from pyramid.security import Allow
-
         permissions = (
+            'pkg_read',
             'pypi_submit',
             'toxresult_upload',
             'index_delete',
@@ -964,7 +1101,8 @@ class PrivateStage(BaseStage):
             ("acl_upload", [self.username]),
             ("acl_toxresult_upload", [":ANONYMOUS:"]),
             ("bases", ()),
-            ("mirror_whitelist", [])]
+            ("mirror_whitelist", []),
+            ("mirror_whitelist_inheritance", "intersection")]
 
     def normalize_indexconfig_value(self, key, value):
         if key == "volatile":
@@ -980,6 +1118,13 @@ class PrivateStage(BaseStage):
             return [
                 normalize_whitelist_name(x)
                 for x in ensure_list(value)]
+        if key == "mirror_whitelist_inheritance":
+            value = value.lower()
+            if value not in ("intersection", "union"):
+                raise InvalidIndexconfig(
+                    "Unknown value '%s' for mirror_whitelist_inheritance, "
+                    "must be 'intersection' or 'union'." % value)
+            return value
         if key in ("custom_data", "description", "title"):
             return value
 
@@ -1086,7 +1231,8 @@ class PrivateStage(BaseStage):
         data = self.key_projsimplelinks(project).get()
         links = data.get("links", [])
         requires_python = data.get("requires_python", [])
-        return join_requires(links, requires_python)
+        yanked = []  # PEP 592 isn't supported for private stages yet
+        return join_links_data(links, requires_python, yanked)
 
     def _regen_simplelinks(self, project_input):
         project = normalize_name(project_input)
@@ -1425,8 +1571,8 @@ class LinkStore:
 
 class SimplelinkMeta(CompareMixin):
     """ helper class to provide information for items from get_simplelinks() """
-    def __init__(self, key_href):
-        self.key, self.href, self.require_python = key_href
+    def __init__(self, link_info):
+        self.key, self.href, self.require_python, self.yanked = link_info
         self._url = URL(self.href)
         self.name, self.version, self.ext = splitbasename(self._url.basename, checkarch=False)
 
@@ -1570,7 +1716,11 @@ class EventSubscribers:
         with keyfs.transaction(at_serial=ev.at_serial) as tx:
 
             if ev.back_serial > -1:
-                old = tx.get_value_at(ev.typedkey, ev.back_serial)
+                try:
+                    old = tx.get_value_at(ev.typedkey, ev.back_serial)
+                except KeyError:
+                    # the user was previously deleted
+                    old = {}
                 old_indexes = set(old.get("indexes", {}))
             else:
                 old_indexes = set()
