@@ -13,28 +13,38 @@ from devpi_common.types import ensure_unicode
 from devpi_common.url import URL
 from devpi_common.metadata import get_pyversion_filetype
 import devpi_server
+from html import escape
 from io import BytesIO
+from lazy import lazy
 from pluggy import HookimplMarker
-from pyramid.compat import escape
-from pyramid.compat import urlparse
-from pyramid.interfaces import IAuthenticationPolicy
+from pyramid.authentication import b64encode
+from pyramid.interfaces import IRequestExtensions
+from pyramid.interfaces import IRootFactory
+from pyramid.interfaces import IRoutesMapper
 from pyramid.httpexceptions import HTTPException, HTTPFound, HTTPSuccessful
+from pyramid.httpexceptions import HTTPForbidden
+from pyramid.httpexceptions import HTTPOk
 from pyramid.httpexceptions import HTTPUnauthorized
 from pyramid.httpexceptions import exception_response
+from pyramid.request import Request
+from pyramid.request import apply_request_extensions
 from pyramid.response import Response
 from pyramid.security import forget
+from pyramid.threadlocal import RequestContext
+from pyramid.traversal import DefaultRootFactory
 from pyramid.view import exception_view_config
 from pyramid.view import view_config
+from urllib.parse import urlparse
 import itertools
 import json
 from devpi_common.request import new_requests_session
 from devpi_common.validation import normalize_name, is_valid_archive_name
 
+from .config import hookimpl
 from .filestore import BadGateway
-from .model import InvalidIndex, InvalidIndexconfig, InvalidUser
+from .model import InvalidIndex, InvalidIndexconfig, InvalidUser, InvalidUserconfig
 from .model import ReadonlyIndex
 from .model import RemoveValue
-from .model import UpstreamError
 from .readonly import get_mutable_deepcopy
 from .log import thread_push_log, thread_pop_log, threadlog
 
@@ -57,6 +67,7 @@ meta_headers = {str("X-DEVPI-API-VERSION"): str(API_VERSION),
 
 
 INSTALLER_USER_AGENT = r"([^ ]* )*(distribute|setuptools|pip|pex)/.*"
+INSTALLER_USER_AGENT_REGEXP = re.compile(INSTALLER_USER_AGENT)
 
 
 def abort(request, code, body):
@@ -117,6 +128,7 @@ def apireturn(code, message=None, result=None, type=None):
     headers = {str("content-type"): str("application/json")}
     raise HTTPResponse(body=data, status=code, headers=headers)
 
+
 def json_preferred(request):
     # XXX do proper "best" matching
     return "application/json" in request.headers.get("Accept", "")
@@ -140,13 +152,13 @@ class OutsideURLMiddleware(object):
         self.xom = xom
 
     def __call__(self, environ, start_response):
-        outside_url = self.xom.config.args.outside_url
+        outside_url = environ.get('HTTP_X_OUTSIDE_URL')
         if not outside_url:
-            outside_url = environ.get('HTTP_X_OUTSIDE_URL')
+            outside_url = self.xom.config.args.outside_url
         if outside_url:
             # XXX memoize it for later access from replica thread
             # self.xom.current_outside_url = outside_url
-            outside_url = urlparse.urlparse(outside_url)
+            outside_url = urlparse(outside_url)
             environ['wsgi.url_scheme'] = outside_url.scheme
             environ['HTTP_HOST'] = outside_url.netloc
             if outside_url.path:
@@ -158,7 +170,7 @@ def route_url(self, *args, **kw):
     url = super(self.__class__, self).route_url(*args, **kw)
     # Unquote plus signs in path segment. The settings in pyramid for
     # the urllib quoting function are a bit too much on the safe side
-    url = urlparse.urlparse(url)
+    url = urlparse(url)
     url = url._replace(path=url.path.replace('%2B', '+'))
     return url.geturl()
 
@@ -189,8 +201,7 @@ def tween_request_logging(handler, registry):
                       duration,
                       serial,
                       rheaders.get("content-length"),
-                      rheaders.get("content-type"),
-            )
+                      rheaders.get("content-type"))
         finally:
             thread_pop_log(tag)
         return response
@@ -295,7 +306,6 @@ class StatusView:
                 seen.add(name)
                 status["metrics"].append(metric)
         if self.xom.is_replica():
-            from .replica import ReplicationErrors
             status["role"] = "REPLICA"
             status["master-url"] = config.master_url.url
             status["master-uuid"] = config.get_master_uuid()
@@ -305,7 +315,7 @@ class StatusView:
             status["master-contacted-at"] = self.xom.replica_thread.master_contacted_at
             status["update-from-master-at"] = self.xom.replica_thread.update_from_master_at
             status["replica-in-sync-at"] = self.xom.replica_thread.replica_in_sync_at
-            replication_errors = ReplicationErrors(self.xom.config.serverdir)
+            replication_errors = self.xom.replica_thread.shared_data.errors
             status["replication-errors"] = replication_errors.errors
         else:
             status["role"] = "MASTER"
@@ -367,6 +377,29 @@ def devpiweb_get_status_info(request):
     return msgs
 
 
+@hookimpl
+def devpiserver_authcheck_always_ok(request):
+    route = request.matched_route
+    if route and route.name.endswith('/+api'):
+        return True
+    if route and route.name == '/+login':
+        return True
+
+
+@hookimpl
+def devpiserver_authcheck_forbidden(request):
+    route = request.matched_route
+    if not route:
+        return
+    route_names = (
+        '/{user}/{index}/+e/{relpath:.*}',
+        '/{user}/{index}/+f/{relpath:.*}')
+    if route.name not in route_names:
+        return
+    if not request.has_permission('pkg_read'):
+        return True
+
+
 class PyPIView:
     def __init__(self, request):
         self.request = request
@@ -374,54 +407,91 @@ class PyPIView:
         xom = request.registry['xom']
         self.xom = xom
         self.model = xom.model
-        self.auth = Auth(self.model, xom.config.secret)
+        self.auth = Auth(self.model, xom.config.get_auth_secret())
         self.log = request.log
 
     def get_auth_status(self):
-        if self.xom.is_replica():
-            from .replica import proxy_request_to_master
-            try:
-                r = proxy_request_to_master(self.xom, self.request)
-                if r.status_code == 200:
-                    return r.json()["result"]["authstatus"]
-            except UpstreamError:
-                pass
-            threadlog.error("could not obtain master authentication status")
-            return ["fail", "", []]
-        # this is accessing some pyramid internals, but they are pretty likely
-        # to stay and the alternative was uglier
-        policy = self.request._get_authentication_policy()
-        credentials = policy._get_credentials(self.request)
-        return self.auth.get_auth_status(credentials)
+        identity = self.request.identity
+        if identity is None:
+            return ["noauth", "", []]
+        return ["ok", identity.username, identity.groups]
 
     #
     # supplying basic API locations for all services
     #
 
     @view_config(route_name="/+api")
-    @view_config(route_name="{path:.*}/+api")
+    @view_config(route_name="/{user}/+api")
+    @view_config(route_name="/{user}/{index}/+api")
     def apiconfig_index(self):
         request = self.request
-        path = request.matchdict.get('path')
+        stage = None
+        if request.context.index is not None:
+            stage = request.context.stage
         api = {
             "login": request.route_url('/+login'),
             "authstatus": self.get_auth_status(),
             "features": self.xom.supported_features,
         }
-        if path:
-            parts = path.split("/")
-            if len(parts) >= 2:
-                user, index = parts[:2]
-                stage = self.context.getstage(user, index)
-                api.update({
-                    "index": request.stage_url(stage),
-                    "simpleindex": request.simpleindex_url(stage)
-                })
-                if stage.ixconfig["type"] != "mirror":
-                    api["pypisubmit"] = request.route_url(
-                        "/{user}/{index}/", user=user, index=index)
+        if stage is not None:
+            api.update({
+                "index": request.stage_url(stage),
+                "simpleindex": request.simpleindex_url(stage)
+            })
+            if stage.ixconfig["type"] != "mirror":
+                api["pypisubmit"] = request.route_url(
+                    "/{user}/{index}/", user=stage.username, index=stage.index)
         apireturn(200, type="apiconfig", result=api)
 
+    def _auth_check_request(self, request):
+        hook = self.xom.config.hook
+        result = hook.devpiserver_authcheck_always_ok(request=request)
+        if result and all(result):
+            threadlog.debug(
+                "Authcheck always OK for %s (%s)",
+                request.url, request.matched_route.name)
+            return HTTPOk()
+        if hook.devpiserver_authcheck_forbidden(request=request):
+            threadlog.debug(
+                "Authcheck Forbidden for %s (%s)",
+                request.url, request.matched_route.name)
+            return HTTPForbidden()
+        if not hook.devpiserver_authcheck_unauthorized(request=request):
+            threadlog.debug(
+                "Authcheck OK for %s (%s)",
+                request.url, request.matched_route.name)
+            return HTTPOk()
+        threadlog.debug(
+            "Authcheck Unauthorized for %s (%s)",
+            request.url, request.matched_route.name)
+        user_agent = request.user_agent or ""
+        if 'devpi-client' in user_agent:
+            # devpi-client needs to know for proper error messages
+            return HTTPForbidden()
+        return HTTPUnauthorized()
+
+    @view_config(route_name="/+authcheck")
+    def authcheck_view(self):
+        request = self.request
+        routes_mapper = request.registry.queryUtility(IRoutesMapper)
+        root_factory = request.registry.queryUtility(
+            IRootFactory, default=DefaultRootFactory)
+        request_extensions = request.registry.queryUtility(IRequestExtensions)
+        url = request.headers.get('x-original-uri', request.url)
+        orig_request = Request.blank(url, headers=request.headers)
+        orig_request.log = request.log
+        orig_request.registry = request.registry
+        if request_extensions:
+            apply_request_extensions(
+                orig_request, extensions=request_extensions)
+        info = routes_mapper(orig_request)
+        (orig_request.matchdict, orig_request.matched_route) = (
+            info['match'], info['route'])
+        root_factory = orig_request.matched_route.factory or root_factory
+        orig_request.context = root_factory(orig_request)
+        with RequestContext(orig_request):
+            return self._auth_check_request(orig_request)
+
     #
     # attach test results to release files
     #
@@ -463,6 +533,12 @@ class PyPIView:
         """
         request = self.request
         abort_if_invalid_project(request, request.matchdict["project"])
+        requested_by_installer = INSTALLER_USER_AGENT_REGEXP.match(
+            request.user_agent or "")
+        if requested_by_installer:
+            # for performance reasons we return results directly for
+            # known installers
+            return self.simple_list_project()
         return HTTPFound(location=self.request.route_url(
             "/{user}/{index}/+simple/{project}/",
             user=self.context.username,
@@ -476,10 +552,10 @@ class PyPIView:
         project = self.context.project
         # we only serve absolute links so we don't care about the route's slash
         stage = self.context.stage
-        requested_by_pip = re.match(INSTALLER_USER_AGENT,
+        requested_by_installer = INSTALLER_USER_AGENT_REGEXP.match(
             request.user_agent or "")
         try:
-            result = stage.get_simplelinks(project, sorted_links=not requested_by_pip)
+            result = stage.get_simplelinks(project, sorted_links=not requested_by_installer)
         except stage.UpstreamError as e:
             threadlog.error(e.msg)
             abort(request, 502, e.msg)
@@ -487,7 +563,7 @@ class PyPIView:
         if not result:
             self.request.context.verified_project  # access will trigger 404 if not found
 
-        if requested_by_pip:
+        if requested_by_installer:
             # we don't need the extra stuff on the simple page for pip
             embed_form = False
             blocked_index = None
@@ -537,22 +613,16 @@ class PyPIView:
             def make_url(href):
                 return url.relpath("/" + href)
 
-        for key, href, require_python in result:
+        for key, href, require_python, yanked in result:
             stage = "/".join(href.split("/", 2)[:2])
-            relurl = make_url(href)
-            if require_python is None:
-                data = dict(stage=stage, url=relurl, key=key)
-                yield ('{stage} <a href="{url}">{key}</a><br/>\n'
-                    .format(**data).encode('utf-8')
-                    )
-            else:
-                require_python = escape(require_python)
-                data = dict(stage=stage, url=relurl, key=key,
-                    require_python=require_python)
-                yield ('{stage} <a data-requires-python="{require_python}" '
-                    'href="{url}">{key}</a><br/>\n'
-                    .format(**data).encode('utf-8')
-                    )
+            attribs = 'href="%s"' % make_url(href)
+            if require_python is not None:
+                attribs += ' data-requires-python="%s"' % escape(require_python)
+            if yanked:
+                attribs += ' data-yanked=""'
+            data = dict(stage=stage, attribs=attribs, key=key)
+            yield '{stage} <a {attribs}>{key}</a><br/>\n'.format(
+                **data).encode('utf-8')
 
         yield "</body></html>".encode("utf-8")
 
@@ -572,7 +642,14 @@ class PyPIView:
         the repository SHOULD redirect the URLs without a /
         to add a / to the end
         """
-        return HTTPFound(location=self.request.route_url(
+        request = self.request
+        requested_by_installer = INSTALLER_USER_AGENT_REGEXP.match(
+            request.user_agent or "")
+        if requested_by_installer:
+            # for performance reasons we return results directly for
+            # known installers
+            return self.simple_list_project()
+        return HTTPFound(location=request.route_url(
             "/{user}/{index}/+simple/",
             user=self.context.username,
             index=self.context.index))
@@ -633,15 +710,12 @@ class PyPIView:
             # we create a user object automatically. The user object may
             # not exist if the user was authenticated by a plugin.
             if self.request.authenticated_userid == username:
-                registry = self.request.registry
-                auth_policy = registry.queryUtility(IAuthenticationPolicy)
-                # we verify the credentials explicitly here, because the
-                # provided token may belong to a deleted user
-                if auth_policy.verify_credentials(self.request):
-                    try:
-                        user = self.model.create_user(username, password=None)
-                    except InvalidUser as e:
-                        apireturn(400, "%s" % e)
+                if not self.request.has_permission("user_create"):
+                    apireturn(403, "no permission to create user %s" % (
+                        self.context.username))
+                user = self.model.create_user(username, password=None)
+                lazy.invalidate(self.context, '_user')
+                lazy.invalidate(self.context, 'user')
         stage = self.context.user.getstage(self.context.index)
         if stage is not None:
             apireturn(409, "index %r exists" % stage.name)
@@ -676,7 +750,7 @@ class PyPIView:
         permission="index_modify")
     def index_modify(self):
         stage = self.context.stage
-        json = _getjson(self.request)
+        json = getjson(self.request)
         if isinstance(json, list):
             ixconfig = stage.get()
             for op, key, value in get_actions(json):
@@ -760,6 +834,9 @@ class PyPIView:
             apireturn(404, message="no release/files found for %s-%s" %(
                       name, version))
 
+        if not request.has_permission("pkg_read"):
+            abort(request, 403, "package read forbidden")
+
         metadata = linkstore.metadata
 
         results = []
@@ -1039,18 +1116,21 @@ class PyPIView:
     #  per-project and version data
     #
 
-    @view_config(route_name="simple_redirect")
-    def simple_redirect(self):
+    @view_config(route_name="installer_simple")
+    def installer_simple(self):
         """
-        PEP 503:
-        the repository SHOULD redirect the URLs without a /
-        to add a / to the end
+        When an installer accesses a project without the /+simple part we
+        return the links directly to avoid a redirect.
         """
-        return HTTPFound(location=self.request.route_url(
-            "/{user}/{index}/+simple/{project}/",
-            user=self.context.username,
-            index=self.context.index,
-            project=self.context.project))
+        return self.simple_list_all()
+
+    @view_config(route_name="installer_simple_project")
+    def installer_simple_project(self):
+        """
+        When an installer accesses a project without the /+simple part we
+        return the links directly to avoid a redirect.
+        """
+        return self.simple_list_project()
 
     @view_config(route_name="/{user}/{index}/{project}",
                  accept="application/json", request_method="GET")
@@ -1074,7 +1154,8 @@ class PyPIView:
     def del_project(self):
         stage = self.context.stage
         project = self.context.project
-        if not stage.ixconfig["volatile"]:
+        force = 'force' in self.request.params
+        if not stage.ixconfig["volatile"] and not force:
             apireturn(403, "project %r is on non-volatile index %s" %(
                       project, stage.name))
         try:
@@ -1117,7 +1198,8 @@ class PyPIView:
     def del_versiondata(self):
         stage = self.context.stage
         name, version = self.context.project, self.context.version
-        if not stage.ixconfig["volatile"]:
+        force = 'force' in self.request.params
+        if not stage.ixconfig["volatile"] and not force:
             abort(self.request, 403, "cannot delete version on non-volatile index")
         try:
             stage.del_versiondata(name, version)
@@ -1153,6 +1235,9 @@ class PyPIView:
                 # redirect to external url
                 return HTTPFound(location=entry.url)
 
+        if not request.has_permission("pkg_read"):
+            abort(request, 403, "package read forbidden")
+
         try:
             if should_fetch_remote_file(entry, request.headers):
                 app_iter = iter_fetch_remote_file(self.xom, entry)
@@ -1198,7 +1283,8 @@ class PyPIView:
                  request_method="DELETE")
     def del_pkg(self):
         stage = self.context.stage
-        if not stage.ixconfig["volatile"]:
+        force = 'force' in self.request.params
+        if not stage.ixconfig["volatile"] and not force:
             abort(self.request, 403, "cannot delete version on non-volatile index")
         relpath = self.request.path_info.strip("/")
         filestore = self.xom.filestore
@@ -1230,13 +1316,24 @@ class PyPIView:
         dict = getjson(request)
         user = dict.get("user", None)
         password = dict.get("password", None)
-        #self.log.debug("got password %r" % password)
         if user is None or password is None:
             abort(request, 400, "Bad request: no user/password specified")
-        proxyauth = self.auth.new_proxy_auth(user, password)
+        proxyauth = self.auth.new_proxy_auth(user, password, request=request)
         if proxyauth:
-            apireturn(200, "login successful", type="proxyauth",
-                result=proxyauth)
+            # set the credentials on the current request
+            request.headers['X-Devpi-Auth'] = b64encode(
+                "%s:%s" % (user, proxyauth['password']))
+            # coherence check of the generated credentials
+            if user != request.authenticated_userid:
+                apireturn(401, "user %r could not be authenticated" % user)
+            # it is possible that a plugin removes the permission to login
+            if not request.has_permission('user_login'):
+                apireturn(
+                    401,
+                    "user %r has no permission to login with the "
+                    "provided credentials" % user)
+            apireturn(
+                200, "login successful", type="proxyauth", result=proxyauth)
         apireturn(401, "user %r could not be authenticated" % user)
 
     @view_config(
@@ -1244,23 +1341,18 @@ class PyPIView:
         permission="user_modify")
     def user_patch(self):
         request = self.request
-        ignored_keys = set(('indexes', 'username'))
-        allowed_keys = set((
-            "email", "password", "title", "description", "custom_data"))
-        result = getjson(request, allowed_keys=allowed_keys.union(ignored_keys))
-        kvdict = dict()
-        for key in allowed_keys:
-            if key not in result:
-                continue
-            kvdict[key] = result[key]
+        kvdict = getjson(request)
         user = self.context.user
         password = kvdict.get("password")
-        user.modify(**kvdict)
+        try:
+            user.modify(**kvdict)
+        except InvalidUserconfig as e:
+            apireturn(400, message=", ".join(e.messages))
         if password is not None:
             apireturn(200, "user updated, new proxy auth",
                       type="userpassword",
-                      result=self.auth.new_proxy_auth(user.name,
-                                                      password=password))
+                      result=self.auth.new_proxy_auth(
+                          user.name, password=password, request=request))
         apireturn(200, "user updated")
 
     @view_config(
@@ -1278,6 +1370,8 @@ class PyPIView:
                 user = self.model.create_user(username, **kvdict)
             except InvalidUser as e:
                 apireturn(400, "%s" % e)
+            except InvalidUserconfig as e:
+                apireturn(400, message=", ".join(e.messages))
             apireturn(201, type="userconfig", result=user.get())
         apireturn(400, "password needs to be set")
 
@@ -1307,9 +1401,6 @@ class PyPIView:
 
     @view_config(route_name="/", accept="application/json", request_method="GET")
     def user_list(self):
-        #accept = request.headers.get("accept")
-        #if accept is not None:
-        #    if accept.endswith("/json"):
         d = {}
         for user in self.model.get_userlist():
             d[user.name] = user.get()
@@ -1322,10 +1413,14 @@ def should_fetch_remote_file(entry, headers):
 
 
 def _headers_from_response(r):
+    content_type = r.headers.get('content-type')
+    if not content_type:
+        content_type = "application/octet-stream"
+    if isinstance(content_type, tuple):
+        content_type = content_type[0]
     headers = {
         str("X-Accel-Buffering"): str("no"),  # disable buffering in nginx
-        str("content-type"): r.headers.get(
-            "content-type", (str("application/octet-stream"), None))[0]}
+        "content-type": content_type}
     if "last-modified" in r.headers:
         headers[str("last-modified")] = r.headers["last-modified"]
     if "content-length" in r.headers:
@@ -1410,17 +1505,22 @@ def iter_cache_remote_file(xom, entry):
 
 
 def iter_remote_file_replica(xom, entry):
-    from .replica import H_REPLICA_FILEREPL, ReplicationErrors
-    replication_errors = ReplicationErrors(xom.config.serverdir)
+    replication_errors = xom.replica_thread.shared_data.errors
     # construct master URL with param
     url = xom.config.master_url.joinpath(entry.relpath).url
     if not entry.url:
         threadlog.warn("missing private file: %s" % entry.relpath)
     else:
         threadlog.info("replica doesn't have file: %s", entry.relpath)
+    (uuid, master_uuid) = make_uuid_headers(xom.config.nodeinfo)
+    rt = xom.replica_thread
+    token = rt.auth_serializer.dumps(uuid)
     r = xom.httpget(
         url, allow_redirects=True,
-        extra_headers={H_REPLICA_FILEREPL: str("YES")})
+        extra_headers={
+            rt.H_REPLICA_FILEREPL: str("YES"),
+            rt.H_REPLICA_UUID: uuid,
+            str('Authorization'): 'Bearer %s' % token})
     if r.status_code != 200:
         msg = "%s: received %s from master" % (url, r.status_code)
         if not entry.url:
@@ -1494,7 +1594,7 @@ def url_for_entrypath(request, entrypath):
         route_name, user=user, index=index, relpath=relpath)
 
 
-def _getjson(request):
+def getjson(request):
     try:
         d = request.json_body
     except ValueError:
@@ -1502,19 +1602,6 @@ def _getjson(request):
     return d
 
 
-def _check_allowed_keys(request, d, allowed_keys):
-    diff = set(d).difference(allowed_keys)
-    if diff:
-        abort(request, 400, "json keys not recognized: %s" % ",".join(diff))
-
-
-def getjson(request, allowed_keys=None):
-    d = _getjson(request)
-    if allowed_keys is not None:
-        _check_allowed_keys(request, d, allowed_keys)
-    return d
-
-
 def abort_if_invalid_filename(request, name, filename):
     if not is_valid_archive_name(filename):
         abort_submit(request, 400, "%r is not a valid archive name" %(filename))
@@ -1523,6 +1610,7 @@ def abort_if_invalid_filename(request, name, filename):
     abort_submit(request, 400, "filename %r does not match project name %r"
                       %(filename, name))
 
+
 def abort_if_invalid_project(request, project):
     try:
         if isinstance(project, bytes):
