@@ -20,6 +20,7 @@ from py.io import BytesIO
 
 pytestmark = [pytest.mark.writetransaction]
 
+
 def udict(**kw):
     """ return a dict where the keys are normalized to unicode. """
     d = {}
@@ -27,10 +28,18 @@ def udict(**kw):
         d[py.builtin._totext(name)] = val
     return d
 
+
 @pytest.fixture(params=[(), ("root/pypi",)])
 def bases(request):
     return request.param
 
+
+def get_release_basenames(stage, project):
+    return sorted(
+        x.basename
+        for x in stage.get_releaselinks(project))
+
+
 def register_and_store(stage, basename, content=b"123", name=None):
     assert py.builtin._isbytes(content), content
     n, version = splitbasename(basename)[:2]
@@ -48,10 +57,12 @@ def stage(request, user):
         config["bases"] = request.getfixturevalue("bases")
     return user.create_stage(**config)
 
+
 @pytest.fixture
 def user(model):
     return model.create_user("hello", password="123")
 
+
 def test_has_mirror_base(model, pypistage):
     pypistage.mock_simple("pytest", "<a href='pytest-1.0.zip' /a>")
     assert pypistage.has_mirror_base("pytest")
@@ -115,6 +126,51 @@ def test_get_mirror_whitelist_info(model, pypistage):
         blocked_by_mirror_whitelist=None)
 
 
+@pytest.mark.notransaction
+def test_get_mirror_whitelist_info_private_package(mapp, monkeypatch, testapp):
+    api = mapp.create_and_use(indexconfig=dict(bases=["root/pypi"]))
+    mapp.set_versiondata(
+        {"name": "pkg1", "version": "2.6", "description": "foo"},
+        set_whitelist=False)
+    # test that getting the whitelist info doesn't reveal package name, see #451
+    with mapp.xom.model.keyfs.transaction(write=False):
+        with monkeypatch.context() as m:
+            # make sure we get an error if data is fetched
+            m.setattr(mapp.xom, 'httpget', None)
+            stage = mapp.xom.model.getstage(api.stagename)
+            info = stage.get_mirror_whitelist_info("pkg1")
+            assert info['has_mirror_base'] is False
+            assert info['blocked_by_mirror_whitelist'] == "root/pypi"
+    mapp.use("root/pypi")
+    mapp.xom.httpget.mock_simple("pkg1", text="")
+    mapp.get_simple("pkg1")
+    with mapp.xom.model.keyfs.transaction(write=False):
+        with monkeypatch.context() as m:
+            # make sure we get an error if data is fetched
+            m.setattr(mapp.xom, 'httpget', None)
+            pypistage = mapp.xom.model.getstage("root/pypi")
+            # expire the project data, otherwise we wouldn't fetch data
+            # by accident
+            pypistage.cache_retrieve_times.expire("pkg1")
+            # now we check that we get correct info without fetching data
+            stage = mapp.xom.model.getstage(api.stagename)
+            info = stage.get_mirror_whitelist_info("pkg1")
+            assert info['has_mirror_base'] is False
+            assert info['blocked_by_mirror_whitelist'] == "root/pypi"
+    # now we whitelist the package
+    testapp.patch_json("/" + api.stagename, ["mirror_whitelist+=pkg1"])
+    with mapp.xom.model.keyfs.transaction(write=False):
+        stage = mapp.xom.model.getstage(api.stagename)
+        info = stage.get_mirror_whitelist_info("pkg1")
+        assert info['has_mirror_base'] is True
+        assert info['blocked_by_mirror_whitelist'] is None
+
+
+@pytest.fixture
+def queue(TimeoutQueue):
+    return TimeoutQueue()
+
+
 class TestStage:
     def test_create_and_delete(self, model):
         user = model.create_user("hello", password="123")
@@ -328,7 +384,7 @@ class TestStage:
         link = register_and_store(stage, "some-1.0.zip", content)
         entry = link.entry
         assert entry.hash_spec
-        assert entry.last_modified != None
+        assert entry.last_modified is not None
         entries = stage.get_releaselinks("some")
         assert len(entries) == 1
         assert entries[0].hash_spec == entry.hash_spec
@@ -405,7 +461,9 @@ class TestStage:
     def test_project_whitelist_inheritance(self, pypistage, stage, user):
         user.create_stage(index="dev2", bases=("root/pypi",))
         stage_dev2 = user.getstage("dev2")
-        stage.modify(bases=(stage_dev2.name,))
+        stage.modify(
+            mirror_whitelist_inheritance="union",
+            bases=(stage_dev2.name,))
         pypistage.mock_simple("someproject",
             "<a href='someproject-1.1.zip' /a>")
         register_and_store(stage, "someproject-1.0.zip", b"123")
@@ -447,7 +505,9 @@ class TestStage:
     def test_project_whitelist_all_inheritance(self, pypistage, stage, user):
         user.create_stage(index="dev2", bases=("root/pypi",))
         stage_dev2 = user.getstage("dev2")
-        stage.modify(bases=(stage_dev2.name,))
+        stage.modify(
+            mirror_whitelist_inheritance="union",
+            bases=(stage_dev2.name,))
         pypistage.mock_simple("someproject",
             "<a href='someproject-1.1.zip' /a>")
         register_and_store(stage, "someproject-1.0.zip", b"123")
@@ -541,6 +601,185 @@ class TestStage:
             lambda self, project: 0 / 0)
         assert stage.list_versions('some_xyz') == {'1.0'}
 
+    def test_whitelist_intersection(self, pypistage, stage, user):
+        pypistage.mock_simple(
+            "someproject", "<a href='someproject-1.1.zip' /a>")
+        stage.modify(mirror_whitelist="*", bases=(pypistage.name,))
+        stage2 = user.create_stage(index='inheriting', bases=(stage.name,))
+        assert stage.ixconfig['mirror_whitelist_inheritance'] == 'intersection'
+        assert stage2.ixconfig['mirror_whitelist_inheritance'] == 'intersection'
+        assert stage2.ixconfig['mirror_whitelist'] == []
+        assert pypistage.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(pypistage, 'someproject') == [
+            'someproject-1.1.zip']
+        assert stage.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage, 'someproject') == [
+            'someproject-1.1.zip']
+        assert stage2.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.1.zip']
+        register_and_store(stage, 'someproject-1.1-py2.py3-none-any.whl')
+        assert stage.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        assert stage2.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl']
+        register_and_store(stage2, 'someproject-1.0.zip')
+        assert stage2.list_versions('someproject') == {'1.0', '1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.0.zip',
+            'someproject-1.1-py2.py3-none-any.whl']
+
+    def test_whitelist_intersection_two_mirrors(self, httpget, stage, user):
+        mirror1 = user.create_stage("mirror1", **udict(
+            mirror_url="http://pypi.org/simple", type="mirror"))
+        mirror2 = user.create_stage("mirror2", **udict(
+            mirror_url="http://example.com/simple", type="mirror"))
+        httpget.mock_simple(
+            "someproject", "<a href='someproject-1.1.zip' /a>",
+            remoteurl=mirror1.mirror_url)
+        httpget.mock_simple(
+            "someproject", "<a href='someproject-1.1-py2.py3-none-any.whl' /a>",
+            remoteurl=mirror2.mirror_url)
+        stage.modify(
+            bases=(mirror1.name, mirror2.name),
+            mirror_whitelist='*')
+        stage2 = user.create_stage(index='inheriting', bases=(stage.name,))
+        assert stage.ixconfig['mirror_whitelist_inheritance'] == 'intersection'
+        assert stage2.ixconfig['mirror_whitelist_inheritance'] == 'intersection'
+        assert mirror1.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(mirror1, 'someproject') == [
+            'someproject-1.1.zip']
+        assert mirror2.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(mirror2, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl']
+        assert stage.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        assert stage2.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        register_and_store(stage, 'someproject-1.0.zip')
+        assert stage.list_versions('someproject') == {'1.0', '1.1'}
+        assert get_release_basenames(stage, 'someproject') == [
+            'someproject-1.0.zip',
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        assert stage2.list_versions('someproject') == {'1.0'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.0.zip']
+
+    def test_whitelist_union(self, pypistage, stage, user):
+        pypistage.mock_simple(
+            "someproject", "<a href='someproject-1.1.zip' /a>")
+        stage.modify(mirror_whitelist="*", bases=(pypistage.name,))
+        stage2 = user.create_stage(
+            index='inheriting', bases=(stage.name,),
+            mirror_whitelist_inheritance='union')
+        assert stage.ixconfig['mirror_whitelist_inheritance'] == 'intersection'
+        assert stage2.ixconfig['mirror_whitelist_inheritance'] == 'union'
+        assert stage2.ixconfig['mirror_whitelist'] == []
+        assert pypistage.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(pypistage, 'someproject') == [
+            'someproject-1.1.zip']
+        assert stage.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage, 'someproject') == [
+            'someproject-1.1.zip']
+        assert stage2.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.1.zip']
+        register_and_store(stage, 'someproject-1.1-py2.py3-none-any.whl')
+        assert stage.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        assert stage2.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        register_and_store(stage2, 'someproject-1.0.zip')
+        assert stage2.list_versions('someproject') == {'1.0', '1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.0.zip',
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+
+    def test_whitelist_union_two_mirrors(self, httpget, stage, user):
+        mirror1 = user.create_stage("mirror1", **udict(
+            mirror_url="http://pypi.org/simple", type="mirror"))
+        mirror2 = user.create_stage("mirror2", **udict(
+            mirror_url="http://example.com/simple", type="mirror"))
+        httpget.mock_simple(
+            "someproject", "<a href='someproject-1.1.zip' /a>",
+            remoteurl=mirror1.mirror_url)
+        httpget.mock_simple(
+            "someproject", "<a href='someproject-1.1-py2.py3-none-any.whl' /a>",
+            remoteurl=mirror2.mirror_url)
+        stage.modify(
+            bases=(mirror1.name, mirror2.name),
+            mirror_whitelist='*')
+        stage2 = user.create_stage(
+            index='inheriting', bases=(stage.name,),
+            mirror_whitelist_inheritance='union')
+        assert stage.ixconfig['mirror_whitelist_inheritance'] == 'intersection'
+        assert stage2.ixconfig['mirror_whitelist_inheritance'] == 'union'
+        assert mirror1.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(mirror1, 'someproject') == [
+            'someproject-1.1.zip']
+        assert mirror2.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(mirror2, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl']
+        assert stage.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        assert stage2.list_versions('someproject') == {'1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        register_and_store(stage, 'someproject-1.0.zip')
+        assert stage.list_versions('someproject') == {'1.0', '1.1'}
+        assert get_release_basenames(stage, 'someproject') == [
+            'someproject-1.0.zip',
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+        assert stage2.list_versions('someproject') == {'1.0', '1.1'}
+        assert get_release_basenames(stage2, 'someproject') == [
+            'someproject-1.0.zip',
+            'someproject-1.1-py2.py3-none-any.whl',
+            'someproject-1.1.zip']
+
+    @pytest.mark.notransaction
+    def test_mirror_whitelist_inheritance_bbb(self, xom):
+        keyfs = xom.keyfs
+        model = xom.model
+        with keyfs.transaction(write=True):
+            user = model.create_user("hello", password="123")
+            config = udict(index="world", bases=(), type="stage", volatile=True)
+            stage = user.create_stage(**config)
+            assert stage.ixconfig["mirror_whitelist_inheritance"] == "intersection"
+            assert stage.get_whitelist_inheritance() == "intersection"
+            with user.key.update() as userconfig:
+                # here we remove the value to simulate an old stage
+                del userconfig['indexes']['world']['mirror_whitelist_inheritance']
+        with keyfs.transaction(write=False):
+            stage = xom.model.getstage("hello/world")
+            assert 'mirror_whitelist_inheritance' not in stage.ixconfig
+            assert stage.get_whitelist_inheritance() == "union"
+        with keyfs.transaction(write=True):
+            stage = xom.model.getstage("hello/world")
+            # now modify an unrelated setting
+            stage.modify(bases=("root/pypi",))
+        with keyfs.transaction(write=False):
+            stage = xom.model.getstage("hello/world")
+            # mirror_whitelist_inheritance should still be missing
+            assert 'mirror_whitelist_inheritance' not in stage.ixconfig
+            assert stage.get_whitelist_inheritance() == "union"
+
     def test_store_and_delete_project(self, stage):
         register_and_store(stage, "some_xyz-1.0.zip", b"123")
         assert stage.get_versiondata_perstage("Some_xyz", "1.0")
@@ -566,16 +805,12 @@ class TestStage:
         assert not stage.has_project_perstage("some-xyz")
 
     def test_delete_not_existing(self, stage, bases):
-        with pytest.raises(stage.NotFound) as excinfo:
+        with pytest.raises(stage.NotFound, match="^project.*not found"):
             stage.del_versiondata("hello", "1.0")
-        assert excinfo.value.msg.startswith("project")
-        assert "not found" in excinfo.value.msg
         register_and_store(stage, "hello-1.0.zip")
         stage.del_versiondata("hello", "1.0", cleanup=False)
-        with pytest.raises(stage.NotFound) as excinfo:
+        with pytest.raises(stage.NotFound, match="^version.*not found"):
             stage.del_versiondata("hello", "1.0")
-        assert excinfo.value.msg.startswith("version")
-        assert "not found" in excinfo.value.msg
 
     def test_releasefile_sorting(self, stage, bases):
         register_and_store(stage, "some-1.1.zip")
@@ -624,7 +859,6 @@ class TestStage:
             archive.extract(tmpdir)
         assert tmpdir.join("index.html").read() == "<html/>"
 
-
     def test_simulate_multiple_doczip_entries(self, stage, bases, tmpdir):
         stage.set_versiondata(udict(name="pkg1", version="1.0"))
         stage.store_doczip("pkg1", "1.0", zip_dict({}))
@@ -653,7 +887,6 @@ class TestStage:
         doczip = stage.get_doczip("pkg1", "1.0")
         assert doczip == content
 
-
     def test_storedoczipfile(self, stage, bases):
         from devpi_common.archive import Archive
         stage.set_versiondata(udict(name="pkg1", version="1.0"))
@@ -851,10 +1084,12 @@ class TestStage:
             def devpiserver_on_upload(self, stage, project, version, link):
                 queue.put((stage, project, version, link))
         stage.xom.config.pluginmanager.register(Plugin())
+
         class Plugin:
             @hookimpl
             def devpiserver_on_remove_file(self, stage, relpath):
                 queue.put((stage, relpath))
+
         stage.xom.config.pluginmanager.register(Plugin())
 
         # upload, should trigger devpiserver_on_upload
@@ -896,9 +1131,8 @@ class TestStage:
         with xom.keyfs.transaction(write=True):
             stage = user.create_stage(**udict(
                 index="world", bases=(), type="stage", volatile=True))
-            with pytest.raises(KeyError) as e:
+            with pytest.raises(KeyError, match='not commited yet'):
                 stage.get_last_change_serial_perstage()
-            assert 'not commited yet' in str(e.value)
         assert current_serial == xom.keyfs.get_current_serial() - 1
         current_serial = xom.keyfs.get_current_serial()
         with xom.keyfs.transaction(write=False):
@@ -1159,10 +1393,6 @@ class TestLinkStore:
 
 class TestUsers:
 
-    def test_secret(self, xom, model):
-        xom.keyfs.basedir.ensure(".something")
-        assert model.get_user(".something") is None
-
     def test_create_and_validate(self, model):
         user = model.get_user("user")
         assert not user
@@ -1174,6 +1404,62 @@ class TestUsers:
         assert user.validate("password")
         assert not user.validate("password2")
 
+    def test_created_and_modified(self, mock, model, monkeypatch):
+        from time import strftime
+        import datetime
+        gmtime_mock = mock.Mock()
+        monkeypatch.setattr("devpi_server.model.gmtime", gmtime_mock)
+        creation_time = datetime.datetime(2021, 2, 22, 10, 51, 51).timetuple()
+        gmtime_mock.return_value = creation_time
+        user = model.create_user("user", "password")
+        assert user
+        userconfig = user.get()
+        assert userconfig["created"] == strftime("%Y-%m-%dT%H:%M:%SZ", creation_time)
+        assert "modified" not in userconfig
+        assert "email" not in userconfig
+        modification_time1 = datetime.datetime(2021, 2, 22, 10, 52, 0).timetuple()
+        gmtime_mock.return_value = modification_time1
+        user.modify(email="some@email.com")
+        userconfig = user.get()
+        assert userconfig["created"] == strftime("%Y-%m-%dT%H:%M:%SZ", creation_time)
+        assert userconfig["modified"] == strftime("%Y-%m-%dT%H:%M:%SZ", modification_time1)
+        assert userconfig["email"] == "some@email.com"
+        # we don't change the email value, but supply created,
+        # which should be ignored
+        gmtime_mock.return_value = datetime.datetime(2021, 2, 22, 10, 52, 1).timetuple()
+        user.modify(created=strftime("%Y-%m-%dT%H:%M:%SZ", modification_time1), email="some@email.com")
+        userconfig = user.get()
+        assert userconfig["created"] == strftime("%Y-%m-%dT%H:%M:%SZ", creation_time)
+        assert userconfig["modified"] == strftime("%Y-%m-%dT%H:%M:%SZ", modification_time1)
+        assert userconfig["email"] == "some@email.com"
+        # we don't change the email value, but supply modified,
+        # which should be ignored
+        gmtime_mock.return_value = datetime.datetime(2021, 2, 22, 10, 52, 1).timetuple()
+        user.modify(modified=strftime("%Y-%m-%dT%H:%M:%SZ", creation_time), email="some@email.com")
+        userconfig = user.get()
+        assert userconfig["created"] == strftime("%Y-%m-%dT%H:%M:%SZ", creation_time)
+        assert userconfig["modified"] == strftime("%Y-%m-%dT%H:%M:%SZ", modification_time1)
+        assert userconfig["email"] == "some@email.com"
+        # we change the email value and supply created and modified
+        # the email should change and modified should be set to gmtime_mock value
+        modification_time2 = datetime.datetime(2021, 2, 22, 10, 52, 1).timetuple()
+        gmtime_mock.return_value = modification_time2
+        user.modify(
+            created=strftime("%Y-%m-%dT%H:%M:%SZ", modification_time1),
+            modified=strftime("%Y-%m-%dT%H:%M:%SZ", creation_time),
+            email="other@email.com")
+        userconfig = user.get()
+        assert userconfig["created"] == strftime("%Y-%m-%dT%H:%M:%SZ", creation_time)
+        assert userconfig["modified"] == strftime("%Y-%m-%dT%H:%M:%SZ", modification_time2)
+        assert userconfig["email"] == "other@email.com"
+        # creating a stage doesn't update modification time
+        assert userconfig["indexes"] == {}
+        user.create_stage("dev")
+        userconfig = user.get()
+        assert userconfig["created"] == strftime("%Y-%m-%dT%H:%M:%SZ", creation_time)
+        assert userconfig["modified"] == strftime("%Y-%m-%dT%H:%M:%SZ", modification_time2)
+        assert "dev" in userconfig["indexes"]
+
     def test_migrate_hash(self, caplog, model):
         from devpi_server.auth import newsalt, getpwhash
         user = model.create_user("user", "password", email="some@email.com")
@@ -1182,18 +1468,17 @@ class TestUsers:
         assert 'pwhash' in userconfig
         salt = newsalt()
         hash = getpwhash("password", salt)
-        user.modify(pwsalt=salt, pwhash=hash)
+        with user.key.update() as userconfig:
+            # modify directly, because the modify method doesn't allow this
+            userconfig["pwsalt"] = salt
+            userconfig["pwhash"] = hash
         userconfig = user.get(credentials=True)
         assert userconfig['pwsalt'] == salt
         assert userconfig['pwhash'] == hash
         # now validate and check for migration
-        recs = caplog.getrecords(".*modified user .*")
-        assert len(recs) == 1
-        assert "pwsalt=*******" in recs[0].getMessage()
         assert user.validate("password")
-        recs = caplog.getrecords(".*modified user .*")
-        assert len(recs) == 2
-        assert "pwsalt=None" in recs[1].getMessage()
+        (rec,) = caplog.getrecords(".*modified user .*")
+        assert "pwsalt=None" in rec.getMessage()
         userconfig = user.get(credentials=True)
         assert 'pwsalt' not in userconfig
         assert userconfig['pwhash'].startswith("$argon2")
@@ -1252,12 +1537,14 @@ class TestUsers:
         user.modify(email=email_address)
         assert user.get()["email"] == email_address
 
+
 def test_user_set_without_indexes(model):
     user = model.create_user("user", "password", email="some@email.com")
     user.create_stage("hello")
     user._set({"password": "pass2"})
     assert model.getstage("user/hello")
 
+
 @pytest.mark.notransaction
 def test_setdefault_indexes(xom, model):
     from devpi_server.main import set_default_indexes
@@ -1380,10 +1667,5 @@ def test_get_indexconfig_values(xom, input, expected):
         with pytest.raises(expected):
             stage.get_indexconfig_from_kwargs(**input)
     else:
-        expected.setdefault("acl_upload", ["user"])
-        expected.setdefault("acl_toxresult_upload", [":ANONYMOUS:"])
-        expected.setdefault("bases", ())
-        expected.setdefault("mirror_whitelist", [])
-        expected.setdefault("volatile", True)
         result = stage.get_indexconfig_from_kwargs(**input)
         assert result == expected
