@@ -16,15 +16,15 @@ def stage(request, model):
 
 
 @pytest.fixture
-def permissionrequest(dummyrequest, model):
-    from devpi_server.view_auth import DevpiAuthenticationPolicy
-    from pyramid.authorization import ACLAuthorizationPolicy
-    from pyramid.interfaces import IAuthenticationPolicy
-    from pyramid.interfaces import IAuthorizationPolicy
-    policy = DevpiAuthenticationPolicy(model.xom)
-    dummyrequest.registry.registerUtility(policy, IAuthenticationPolicy)
-    policy = ACLAuthorizationPolicy()
-    dummyrequest.registry.registerUtility(policy, IAuthorizationPolicy)
+def permissionrequest(model):
+    from devpi_server.view_auth import DevpiSecurityPolicy
+    from pyramid.interfaces import ISecurityPolicy
+    from pyramid.request import Request
+    from pyramid.threadlocal import get_current_registry
+    policy = DevpiSecurityPolicy(model.xom)
+    dummyrequest = Request.blank('/')
+    dummyrequest.registry = get_current_registry()
+    dummyrequest.registry.registerUtility(policy, ISecurityPolicy)
     dummyrequest.log = model.xom.log
     return dummyrequest
 
@@ -33,12 +33,12 @@ def permissionrequest(dummyrequest, model):
 def plugin():
     class Plugin:
         @hookimpl
-        def devpiserver_auth_user(self, userdict, username, password):
+        def devpiserver_auth_request(self, request, userdict, username, password):
             if username == 'external':
                 return dict(
                     status='ok',
                     groups=['developer'])
-            return dict(status="unknown")
+            return None
     return Plugin()
 
 
@@ -49,6 +49,7 @@ def xom(makexom, plugin):
 
 def with_user(request, user):
     from pyramid.authentication import b64encode
+    from pyramid.security import ISecurityPolicy
     if user is None:
         request.headers.pop('Authorization', None)
     else:
@@ -58,6 +59,8 @@ def with_user(request, user):
             auth = "%s:123" % user
         request.headers['Authorization'] = 'Basic %s' % b64encode(
             auth.encode('utf-8')).decode('ascii')
+    policy = request.registry.queryUtility(ISecurityPolicy)
+    policy.identity_cache.clear(request)
     return request
 
 
@@ -105,7 +108,7 @@ class TestStage:
         assert indexconfig["acl_toxresult_upload"] == [":ANONYMOUS:"]
 
         stage = model.getstage("hello/world")
-        
+
         # anonymous may upload tests
         assert with_user(permissionrequest, None).has_permission(
             'toxresult_upload', stage)
@@ -139,3 +142,37 @@ class TestStage:
         # except external can upload
         assert with_user(permissionrequest, 'external').has_permission(
             'toxresult_upload', stage)
+
+
+class TestAuthDenialPlugin:
+    @pytest.fixture
+    def plugin(self):
+        class Plugin:
+            @hookimpl
+            def devpiserver_auth_denials(self, request, acl, user, stage):
+                return self.results.pop()
+        return Plugin()
+
+    @pytest.fixture
+    def xom(self, makexom, plugin):
+        xom = makexom(plugins=[plugin])
+        return xom
+
+    def test_denials_plugin(self, permissionrequest, plugin, xom):
+        from devpi_server.view_auth import RootFactory
+        request = with_user(permissionrequest, 'root')
+        request.registry['xom'] = xom
+        context = RootFactory(request)
+        plugin.results = [None]
+        assert request.has_permission('user_create', context)
+        context = RootFactory(request)
+        plugin.results = [[('root', 'user_create')]]
+        assert not request.has_permission('user_create', context)
+
+    @pytest.mark.notransaction
+    def test_deny_login(self, plugin, xom, mapp):
+        plugin.results = [None]
+        mapp.login("root", "")
+        mapp.logout()
+        plugin.results = [[('root', 'user_login')]]
+        mapp.login("root", "", code=401)
