@@ -91,18 +91,30 @@ def test_user_patch_keeps_missing_keys(testapp):
     testapp.put_json("/foo", dict(password="123"))
     testapp.set_auth('foo', '123')
     r = testapp.get("/foo")
-    assert r.json['result'] == {'username': 'foo', 'indexes': {}}
+    result = r.json['result']
+    result.pop('created', None)
+    result.pop('modified', None)
+    assert result == {'username': 'foo', 'indexes': {}}
     testapp.patch_json("/foo", dict(title="foo"))
     r = testapp.get("/foo")
-    assert r.json['result'] == {
+    result = r.json['result']
+    result.pop('created', None)
+    result.pop('modified', None)
+    assert result == {
         'username': 'foo', 'title': 'foo', 'indexes': {}}
     testapp.patch_json("/foo", dict(description="bar"))
     r = testapp.get("/foo")
-    assert r.json['result'] == {
+    result = r.json['result']
+    result.pop('created', None)
+    result.pop('modified', None)
+    assert result == {
         'username': 'foo', 'title': 'foo', 'description': 'bar', 'indexes': {}}
     testapp.patch_json("/foo", dict(description=""))
     r = testapp.get("/foo")
-    assert r.json['result'] == {
+    result = r.json['result']
+    result.pop('created', None)
+    result.pop('modified', None)
+    assert result == {
         'username': 'foo', 'title': 'foo', 'indexes': {}}
 
 
@@ -116,6 +128,7 @@ def test_make_uuid_headers(nodeinfo, expected):
     output = make_uuid_headers(nodeinfo)
     assert output == expected
 
+
 def test_simple_project(pypistage, testapp):
     name = "qpwoei"
     r = testapp.get("/root/pypi/+simple/%s/" % name)
@@ -132,7 +145,8 @@ def test_simple_project(pypistage, testapp):
     links = getlinks(r.text)
     assert len(links) == 1
     assert links[0].get("href").endswith(path)
-    assert links[0].get("data-requires-python") == None
+    assert links[0].get("data-requires-python") is None
+
 
 def test_simple_project_requires_python(pypistage, testapp):
     name = "django"
@@ -145,6 +159,7 @@ def test_simple_project_requires_python(pypistage, testapp):
     assert links[0].get("href").endswith(path)
     assert links[0].get("data-requires-python") == '>=3.4'
 
+
 def test_simple_project_requires_python_empty(pypistage, testapp):
     name = "django"
     path = "/%s-2.0.tar.gz" % name
@@ -156,17 +171,32 @@ def test_simple_project_requires_python_empty(pypistage, testapp):
     assert links[0].get("href").endswith(path)
     assert links[0].get("data-requires-python") is None
 
+
+def test_simple_project_yanked(pypistage, testapp):
+    name = "django"
+    path = "/%s-2.0.tar.gz" % name
+    pypistage.mock_simple(name, text='<a href="%s" data-yanked="" />' % path)
+    r = testapp.get("/root/pypi/+simple/%s/" % name)
+    assert r.status_code == 200
+    links = BeautifulSoup(r.text, "html.parser").findAll("a")
+    assert len(links) == 1
+    assert links[0].get("href").endswith(path)
+    assert links[0].get("data-yanked") == ""
+
+
 def test_simple_list_all_redirect(pypistage, testapp):
     r = testapp.get("/root/pypi/+simple", follow=False)
     assert r.status_code == 302
     assert r.headers["location"].endswith("/root/pypi/+simple/")
 
+
 def test_simple_project_redirect(pypistage, testapp):
     project = "py"
     r = testapp.get("/root/pypi/+simple/" + project, follow=False)
     assert r.status_code == 302
     assert r.headers["location"].endswith("/root/pypi/+simple/%s/" % project)
 
+
 @pytest.mark.parametrize("outside_url", ['', 'http://localhost/devpi'])
 def test_simple_project_outside_url_subpath(mapp, outside_url, pypistage, testapp):
     api = mapp.create_and_use(indexconfig=dict(bases=["root/pypi"]))
@@ -225,7 +255,7 @@ def test_simple_project_absolute_url(mapp, pypistage, testapp):
         headers=headers)
 
 
-@pytest.mark.parametrize(
+user_agent_parameter_args = (
     "user_agent",
     [
         'pip/1.4.1',
@@ -233,23 +263,102 @@ def test_simple_project_absolute_url(mapp, pypistage, testapp):
         'Python-urllib/3.5 setuptools/6.1',
         'setuptools/6.1 Python-urllib/3.5',
         'pex/1.0.1',
-        'pip/6.0.dev1 {"cpu":"x86_64","distro":{"name":"OS X","version":"10.9.5"},"implementation":{"name":"CPython","version":"2.7.8"},"installer":{"name":"pip","version":"6.0.dev1"},"python":"2.7.8","system":{"name":"Darwin","release":"13.4.0"}}'],
+        'pip/6.0.dev1 {"cpu":"x86_64","distro":{"name":"OS X","version":"10.9.5"},"implementation":{"name":"CPython","version":"2.7.8"},"installer":{"name":"pip","version":"6.0.dev1"},"python":"2.7.8","system":{"name":"Darwin","release":"13.4.0"}}'])
+user_agent_parameter_kw = dict(
     ids=['pip', 'setuptools', 'urllib-setuptools', 'setuptools-urllib', 'pex', 'pip6'])
-def test_project_redirect(pypistage, testapp, user_agent):
-    name = "qpwoei"
+
+
+@pytest.mark.nomockprojectsremote
+@pytest.mark.parametrize(*user_agent_parameter_args, **user_agent_parameter_kw)
+def test_projects_simple_results_for_installers(pypistage, testapp, user_agent):
+    # installers should get the simple projects info directly without redirect
+    pypistage.mock_simple_projects(["hello1", "hello2"])
+    headers = {'User-Agent': str(user_agent), "Accept": str("text/html")}
+    r = testapp.get("/root/pypi", headers=headers, follow=False)
+    assert r.status_code == 200
+    assert 'href="hello1/"' in r.text
+    assert 'href="hello2/"' in r.text
+    r = testapp.get("/root/pypi/", headers=headers, follow=False)
+    assert r.status_code == 200
+    assert 'href="hello1/"' in r.text
+    assert 'href="hello2/"' in r.text
+
+
+def test_projects_redirect(pypistage, testapp):
+    # test non installer html requests which should go nowhere,
+    # because devpi-server only returns json
+    headers = {"Accept": str("text/html")}
+    r = testapp.get("/root/pypi", headers=headers, follow=False)
+    assert r.status_code == 404
+    r = testapp.get("/root/pypi/", headers=headers, follow=False)
+    assert r.status_code == 404
+
+
+@pytest.mark.parametrize(*user_agent_parameter_args, **user_agent_parameter_kw)
+def test_project_simple_results_for_installers(pypistage, testapp, user_agent):
+    import os.path
+    # installers should get the simple page info directly without redirect
+    pypistage.mock_simple(
+        "py",
+        """<a href="/py-2.0.zip" /><a href="/py-3.0.zip" /><a href="/py-1.0.zip" />""")
     headers = {'User-Agent': str(user_agent), "Accept": str("text/html")}
 
+    r = testapp.get("/root/pypi/py", headers=headers, follow=False)
+    assert r.status_code == 200
+    links = [os.path.basename(x.attrs["href"]) for x in getlinks(r.text)]
+    assert links == ["py-2.0.zip", "py-3.0.zip", "py-1.0.zip"]
+    r = testapp.get("/root/pypi/py/", headers=headers, follow=False)
+    assert r.status_code == 200
+    links = [os.path.basename(x.attrs["href"]) for x in getlinks(r.text)]
+    assert links == ["py-2.0.zip", "py-3.0.zip", "py-1.0.zip"]
+    r = testapp.get("/root/pypi/+simple/py", headers=headers, follow=False)
+    assert r.status_code == 200
+    links = [os.path.basename(x.attrs["href"]) for x in getlinks(r.text)]
+    assert links == ["py-2.0.zip", "py-3.0.zip", "py-1.0.zip"]
+
+
+def test_project_redirect(pypistage, testapp):
+    name = "qpwoei"
+    # test non installer html requests which should go nowhere,
+    # because devpi-server only returns json
+    headers = {"Accept": str("text/html")}
+
     r = testapp.get("/root/pypi/%s" % name, headers=headers, follow=False)
-    assert r.status_code == 302
-    assert r.headers["location"].endswith("/root/pypi/+simple/%s/" % name)
+    assert r.status_code == 404
     r = testapp.get("/root/pypi/%s/" % name, headers=headers, follow=False)
-    assert r.status_code == 302
-    assert r.headers["location"].endswith("/root/pypi/+simple/%s/" % name)
-    # check with x-outside-url
-    headers['X-Outside-Url'] = str('http://example.com/path')
-    r = testapp.get("/root/pypi/%s/" % name, headers=headers, follow=False)
-    assert r.status_code == 302
-    assert r.headers["location"] == "http://example.com/path/root/pypi/+simple/%s/" % name
+    assert r.status_code == 404
+
+
+@pytest.mark.parametrize(*user_agent_parameter_args, **user_agent_parameter_kw)
+def test_simple_project_plain_info_for_installers(monkeypatch, pypistage, testapp, user_agent):
+    from devpi_server.model import BaseStage
+    import os.path
+    pypistage.mock_simple(
+        "py",
+        """<a href="/py-2.0.zip" /><a href="/py-3.0.zip" /><a href="/py-1.0.zip" />""")
+    headers = {'User-Agent': str(user_agent), "Accept": str("text/html")}
+    # fail if called
+    monkeypatch.setattr(
+        BaseStage, "get_mirror_whitelist_info", lambda s, p: 0 / 0)
+    orig_get_simplelinks = BaseStage.get_simplelinks
+    calls = []
+
+    def get_simplelinks(self, project, sorted_links=True):
+        calls.append((self, project, sorted_links))
+        return orig_get_simplelinks(self, project, sorted_links=sorted_links)
+
+    # let original be called, but we can inspect the call now
+    monkeypatch.setattr(BaseStage, "get_simplelinks", get_simplelinks)
+
+    r = testapp.get("/root/pypi/+simple/py/", headers=headers)
+    ((stage, name, sorted_links),) = calls
+    links = [os.path.basename(x.attrs["href"]) for x in getlinks(r.text)]
+    assert links == ["py-2.0.zip", "py-3.0.zip", "py-1.0.zip"]
+    assert stage.name == pypistage.name
+    assert name == "py"
+    assert sorted_links is False
+    assert "<form" not in r.text
+    assert "<input" not in r.text
 
 
 def test_simple_project_unicode_rejected(pypistage, testapp, dummyrequest):
@@ -265,10 +374,12 @@ def test_simple_project_unicode_rejected(pypistage, testapp, dummyrequest):
     with pytest.raises(HTTPClientError):
         view.simple_list_project()
 
+
 def test_simple_url_longer_triggers_404(testapp):
     assert testapp.get("/root/pypi/+simple/pytest/1.0/").status_code == 404
     assert testapp.get("/root/pypi/+simple/pytest/1.0").status_code == 404
 
+
 def test_simple_project_pypi_egg(pypistage, testapp):
     pypistage.mock_simple("py",
         """<a href="http://bb.org/download/py.zip#egg=py-dev" />""")
@@ -279,6 +390,7 @@ def test_simple_project_pypi_egg(pypistage, testapp):
     r = testapp.get("/root/pypi")
     assert r.status_code == 200
 
+
 @pytest.mark.nomockprojectsremote
 def test_simple_list(pypistage, testapp):
     pypistage.mock_simple_projects(["hello1", "hello2"])
@@ -302,6 +414,7 @@ def test_simple_list(pypistage, testapp):
     hrefs = [a.get("href") for a in links]
     assert hrefs == ["hello1/", "hello2/"]
 
+
 def test_correct_resolution_order(pypistage, mapp, testapp):
     pypistage.mock_simple("hello", pkgver="hello-1.0.tar.gz")
     index1 = mapp.create_and_use()
@@ -316,6 +429,7 @@ def test_correct_resolution_order(pypistage, mapp, testapp):
     r = testapp.get("/%s/+simple/hello" % index3.stagename)
     assert index1.stagename in r.text
 
+
 def test_simple_page_pypi_serial(pypistage, testapp):
     # test with no header
     pypistage.mock_simple("hello1", text="qwe", pypiserial=None)
@@ -331,6 +445,7 @@ def test_simple_page_pypi_serial(pypistage, testapp):
     r = testapp.get("/root/pypi/+simple/hello3/", expect_errors=False)
     assert "X-PYPI-LAST-SERIAL" not in r.headers
 
+
 def test_simple_refresh(mapp, model, pypistage, testapp):
     pypistage.mock_simple("hello", "<html/>")
     r = testapp.xget(200, "/root/pypi/+simple/hello/")
@@ -539,12 +654,13 @@ def test_pkgserv_remote_failure(httpget, pypistage, testapp):
     r = testapp.get(url)
     assert r.status_code == 502
 
+
 def test_apiconfig(testapp):
     r = testapp.get_json("/user/name/+api", status=404)
     assert r.status_code == 404
     r = testapp.get_json("/root/pypi/+api")
     assert r.status_code == 200
-    assert not "pypisubmit" in r.json["result"]
+    assert "pypisubmit" not in r.json["result"]
 
 
 def test_apiconfig_features(testapp):
@@ -582,14 +698,12 @@ def test_apiconfig_with_outside_url(testapp):
     assert result["login"] == u + "/+login"
     assert result["simpleindex"] == u + "/root/pypi/+simple/"
 
-    #for name in "pushrelease simpleindex login pypisubmit resultlog".split():
-    #    assert name in r.json
-    #
-    #
+
 def test_root_pypi(testapp):
     r = testapp.get("/root/pypi")
     assert r.status_code == 200
 
+
 def test_set_versiondata_and_get_description(mapp, testapp):
     api = mapp.create_and_use("user/name")
     metadata = {"name": "pkg1", "version": "1.0", ":action": "submit",
@@ -607,6 +721,7 @@ def test_set_versiondata_and_get_description(mapp, testapp):
     assert r.status_code == 200
     assert "1.0" in r.json["result"]
 
+
 class TestSubmitValidation:
     @pytest.fixture
     def submit(self, mapp, testapp):
@@ -1009,6 +1124,7 @@ def test_push_non_existent(mapp, testapp, monkeypatch):
     r = testapp.push("/user1/dev", json.dumps(req), expect_errors=True)
     assert r.status_code == 401
 
+
 def test_push_from_base_error(mapp, testapp, monkeypatch, pypistage):
     pypistage.mock_simple("hello", text='<a href="hello-1.0.tar.gz"/>')
     mapp.create_and_login_user("user1", "1")
@@ -1235,6 +1351,7 @@ def test_upload_and_delete_project(mapp, testapp):
     mapp.getjson(api.index + "/pkg1", code=404)
     mapp.getjson(api.index + "/pkg1/2.7", code=404)
 
+
 def test_upload_with_acl(mapp):
     mapp.login("root")
     mapp.change_password("root", "123")
@@ -1274,11 +1391,13 @@ class TestPluginPermissions:
     def plugin(self):
         class Plugin:
             groups = ['plugingroup']
+
             @hookimpl
-            def devpiserver_auth_user(self, userdict, username, password):
-                if username == 'pluginuser':
+            def devpiserver_auth_request(self, request, userdict, username, password):
+                if username == 'pluginuser' and password == 'pluginpassword':
                     return dict(status="ok", groups=self.groups)
-                return dict(status="unknown")
+                return None
+
         return Plugin()
 
     @pytest.fixture
@@ -1290,13 +1409,13 @@ class TestPluginPermissions:
         mapp.login("root")
         mapp.create_and_use()  # new context and login
         mapp.set_versiondata(dict(name="pkg1", version="1.0"))
-        mapp.login("pluginuser")
+        mapp.login("pluginuser", "pluginpassword")
         # pluginuser cannot write to index now
         mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=403)
         # now we change the acl
         mapp.login("root")
         mapp.set_acl([":plugingroup"])
-        mapp.login("pluginuser")
+        mapp.login("pluginuser", "pluginpassword")
         # we need to skip setting the whitelist here, because the user may only
         # register and upload a package, but not modify the index
         mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6",
@@ -1304,11 +1423,11 @@ class TestPluginPermissions:
         # if we remove the user from the group (and login again, as the groups
         # are stored in the token) she can't upload anymore
         plugin.groups = []
-        mapp.login("pluginuser")
+        mapp.login("pluginuser", "pluginpassword")
         mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=403)
 
     def test_plugin_user_create_index(self, mapp):
-        mapp.login("pluginuser")
+        mapp.login("pluginuser", "pluginpassword")
         assert "pluginuser" not in mapp.getuserlist()
         mapp.create_index("pluginuser/dev")
         assert "pluginuser" in mapp.getuserlist()
@@ -1358,6 +1477,7 @@ def test_upload_and_access_releasefile_meta(mapp, testapp, proj):
     hash_spec = pkgmeta["result"]["hash_spec"]
     assert hash_spec_matches(hash_spec, content)
 
+
 def test_upload_and_delete_project_version(mapp):
     api = mapp.create_and_use()
     mapp.delete_project("pkg1", code=404)
@@ -1371,12 +1491,20 @@ def test_upload_and_delete_project_version(mapp):
     #assert mapp.getjson("/user/name/pkg1/")["status"] == 404
     mapp.getjson(api.index + "pkg1", code=404)
 
+
 def test_delete_version_fails_on_non_volatile(mapp):
     mapp.create_and_use(indexconfig=dict(volatile=False))
     mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6")
     mapp.delete_project("pkg1/2.6", code=403)
 
 
+def test_delete_version_on_non_volatile_force(mapp):
+    mapp.create_and_use(indexconfig=dict(volatile=False))
+    mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6")
+    mapp.delete_project("pkg1/2.6", code=403)
+    mapp.delete_project("pkg1/2.6?force")
+
+
 @pytest.mark.nomocking
 def test_upload_to_mirror_fails(mapp, simpypi):
     indexconfig = dict(
@@ -1569,6 +1697,17 @@ def test_delete_volatile_fails(mapp):
     mapp.use("root/test")
     mapp.upload_file_pypi("pkg5-2.6.tgz", b"123", "pkg5", "2.6")
     mapp.delete_project("pkg5", code=403)
+    assert mapp.getpkglist() == ["pkg5"]
+
+
+def test_delete_volatile_force(mapp, testapp):
+    mapp.login_root()
+    mapp.create_index("test", indexconfig=dict(volatile=False))
+    api = mapp.use("root/test")
+    mapp.upload_file_pypi("pkg5-2.6.tgz", b"123", "pkg5", "2.6")
+    assert mapp.getpkglist() == ["pkg5"]
+    testapp.delete_json("/%s/pkg5?force" % api.stagename)
+    assert mapp.getpkglist() == []
 
 
 @pytest.mark.nomocking
@@ -1633,6 +1772,26 @@ def test_delete_with_acl_upload(mapp, restrict_modify, volatile, xom):
     mapp.delete_project('pkg5', code=result_code)
 
 
+def test_delete_root(mapp, xom):
+    mapp.login_root()
+    # we need to set the index to non-volatile, otherwise that will
+    # trigger the 403 instead of the missing user_delete permission
+    res = mapp.getjson("/root/pypi")["result"]
+    mapp.modify_index("root/pypi", indexconfig=dict(res, volatile=True))
+    mapp.delete_user("root", code=403)
+
+
+def test_delete_self_restrict_modify(mapp, xom):
+    xom.config.args.restrict_modify = "root,admin"
+    mapp.login("root")
+    mapp.create_user("admin", "")
+    mapp.login("admin")
+    mapp.delete_user("admin", code=403)
+    mapp.login("root")
+    mapp.delete_user("root", code=403)
+    mapp.delete_user("admin")
+
+
 def test_delete_package(mapp, testapp):
     mapp.login_root()
     mapp.create_index("test")
@@ -1814,6 +1973,16 @@ def test_delete_package_volatile_fails(mapp, testapp):
     testapp.xdel(403, link.href)
 
 
+def test_delete_package_volatile_force(mapp, testapp):
+    mapp.login_root()
+    mapp.create_index("test", indexconfig=dict(volatile=False))
+    mapp.use("root/test")
+    mapp.upload_file_pypi("pkg5-2.6.tgz", b"123", "pkg5", "2.6")
+    vv = get_view_version_links(testapp, "/root/test", "pkg5", "2.6")
+    (link,) = vv.get_links()
+    testapp.xdel(200, link.href + '?force')
+
+
 @proj
 def test_upload_docs_no_version(mapp, testapp, proj):
     api = mapp.create_and_use()
@@ -1907,13 +2076,14 @@ def test_wrong_login_format(testapp, mapp):
     (
         {"Host": "outside3.com:3141"}, {'wsgi.url_scheme': 'https'},
         None, "https://outside3.com:3141"),
-    # outside url takes precedence over headers
+    # x-outside-url header takes precedence over outside url option
     (
         {"X-outside-url": "http://outside.com"}, {},
-        "http://outside2.com", "http://outside2.com"),
+        "http://outside2.com", "http://outside.com"),
     (
         {"X-outside-url": "http://outside.com"}, {},
-        "http://outside2.com/foo", "http://outside2.com/foo"),
+        "http://outside2.com/foo", "http://outside.com"),
+    # outside url option takes precedence over host and environ
     (
         {"Host": "outside3.com"}, {},
         "http://out.com", "http://out.com"),
@@ -1993,9 +2163,11 @@ class Test_getjson:
     @pytest.fixture
     def abort_calls(self, monkeypatch):
         l = []
+
         def recorder(*args, **kwargs):
             l.append((args, kwargs))
             raise SystemExit(1)
+
         monkeypatch.setattr(devpi_server.views, "abort", recorder)
         return l
 
@@ -2015,42 +2187,39 @@ class Test_getjson:
         abort_call_args = abort_calls[0][0]
         assert abort_call_args[1] == 400
 
-    def test_getjson_wrong_keys(self, abort_calls):
-        from devpi_server.views import getjson
-        from pyramid.request import Request
-        request = Request({}, body=b'{"k1": "v1", "k2": "v2"')
-        with pytest.raises(SystemExit):
-            getjson(request, allowed_keys=["k1", "k3"])
-        assert len(abort_calls) == 1
-        abort_call_args = abort_calls[0][0]
-        assert abort_call_args[1] == 400
-
 
 class TestTweenKeyfsTransaction:
     def test_nowrite(self, xom, blank_request):
         cur_serial = xom.keyfs.get_current_serial()
-        wrapped_handler = lambda r: Response("")
+
+        def wrapped_handler(r):
+            return Response("")
+
         handler = tween_keyfs_transaction(wrapped_handler, {"xom": xom})
         response = handler(blank_request())
         assert response.headers.get("X-DEVPI-SERIAL") == str(cur_serial)
 
     def test_write(self, xom, blank_request):
         cur_serial = xom.keyfs.get_current_serial()
+
         def wrapped_handler(request):
             with xom.keyfs.USER(user="hello").update():
                 pass
             return Response("")
+
         handler = tween_keyfs_transaction(wrapped_handler, {"xom": xom})
         response = handler(blank_request(method="PUT"))
         assert response.headers.get("X-DEVPI-SERIAL") == str(cur_serial + 1)
 
     def test_restart(self, xom, blank_request):
         cur_serial = xom.keyfs.get_current_serial()
+
         def wrapped_handler(request):
             xom.keyfs.restart_as_write_transaction()
             with xom.keyfs.USER(user="hello").update():
                 pass
             return Response("")
+
         handler = tween_keyfs_transaction(wrapped_handler, {"xom": xom})
         response = handler(blank_request())
         assert response.headers.get("X-DEVPI-SERIAL") == str(cur_serial + 1)
@@ -2064,12 +2233,12 @@ class TestRestrictModify:
     def plugin(self):
         class Plugin:
             @hookimpl
-            def devpiserver_auth_user(self, userdict, username, password):
+            def devpiserver_auth_request(self, request, userdict, username, password):
                 if username == "regular" and password == "regular":
                     return dict(status="ok", groups=["regulars"])
                 if username == "admin" and password == "admin":
                     return dict(status="ok", groups=["admins"])
-                return dict(status="unknown")
+                return None
         return Plugin()
 
     @pytest.fixture
